----------------------------------------------------------------------------------
--
-- LGP-30 drum computer simulation
-- Magnetic drum module (main memory, re-circulating registers, sync signals)
--
-- jm 31.7.2016
-- jm 16.10.2017	register cache for HDMI output is updated from ARC write signals (> immediate screen update) 
--
-- Internal timing is different from the LGP-30:
-- LGP-30 has read flip-flops, but direkt write-through.
-- Simulation memory is latched (clocked) on read and write access.
-- To compensate, the write addresses are always lagging one position behind what they are in the LGP-30.
-- This also means a multiplexed read/write address for the main memory, where the write address is lagging by 1.
--
-- Dual ported memory is used, with the second port reserved for access by the HDMI video output.
--
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity drum is
	port (
		
		-- system clock
		clk: in std_logic;
		
		-- main memory
		V, Vn: out std_logic;					-- read head
		Vp: in std_logic;							-- write head
		W: in std_logic;							-- write enable
		P: in std_logic_vector (1 to 6);		-- track selector
	
		-- recycling registers
		A, An: out std_logic;					-- accumulator read, 32 bits behind write
		As, Asn: out std_logic;					-- accumulator read, 65 bits behind write
		Ap: in std_logic;							-- accumulator write, always enabled

		R, Rn: out std_logic;					-- instruction read, 32 bits behind write
		Rp: in std_logic;							-- instruction write, always enabled

		C, Cn: out std_logic;					-- program counter read, 32 bits behind write
		Cp: in std_logic;							-- program counter write, always enabled
		
		-- timing signals
		S1, S1n: out std_logic;
		S2, S2n: out std_logic;
		S3, S3n: out std_logic;
		
		-- memory access for HDMI display (main drum memory)
		vid_clk: in std_logic;
		vid_adr: in std_logic_vector (16 downto 0);
		vid_data: out std_logic;
		vid_regs: out std_logic_vector (5 downto 0);	-- S3, S2, S1, A, R, C
		drumpos: out std_logic_vector (10 downto 0);	-- current drum rotary position
		
		-- control for xmodem-driven memory access
		sec_reset: in std_logic								-- async. reset of rotary position to sector 0
	);
end drum;

architecture Behavioral of drum is

	-- recirculating registers (one each for A, R, C, plus one for As)
	-- dual port - always written, always read via displaced read head
	COMPONENT circulate
		PORT (
			clka : IN STD_LOGIC;
			wea : IN STD_LOGIC_VECTOR(0 DOWNTO 0);
			addra : IN STD_LOGIC_VECTOR(10 DOWNTO 0);
			dina : IN STD_LOGIC_VECTOR(0 DOWNTO 0);
			clkb : IN STD_LOGIC;
			addrb : IN STD_LOGIC_VECTOR(10 DOWNTO 0);
			doutb : OUT STD_LOGIC_VECTOR(0 DOWNTO 0)
		);
	END COMPONENT;
	
	-- main memory - 64 * 64 * 32 bits
	-- dual port, port B is for access by the HDMI display.
	-- Port A (LGP-30) either reads or writes in any given cycle, but read and write address must be displaced by one.
	COMPONENT memory
	  PORT (
		 clka : IN STD_LOGIC;
		 wea : IN STD_LOGIC_VECTOR(0 DOWNTO 0);
		 addra : IN STD_LOGIC_VECTOR(16 DOWNTO 0);
		 dina : IN STD_LOGIC_VECTOR(0 DOWNTO 0);
		 douta : OUT STD_LOGIC_VECTOR(0 DOWNTO 0);
		 clkb : IN STD_LOGIC;
		 web : IN STD_LOGIC_VECTOR(0 DOWNTO 0);
		 addrb : IN STD_LOGIC_VECTOR(16 DOWNTO 0);
		 dinb : IN STD_LOGIC_VECTOR(0 DOWNTO 0);
		 doutb : OUT STD_LOGIC_VECTOR(0 DOWNTO 0)
	  );
	END COMPONENT;

	-- copy of registers and sync tracks for video output.
	-- dual port memory; input A,R,C,S1-S3; output to HDMI unit
	COMPONENT cache
	  PORT (
		 clka : IN STD_LOGIC;
		 wea : IN STD_LOGIC_VECTOR(0 DOWNTO 0);
		 addra : IN STD_LOGIC_VECTOR(10 DOWNTO 0);
		 dina : IN STD_LOGIC_VECTOR(5 DOWNTO 0);
		 clkb : IN STD_LOGIC;
		 addrb : IN STD_LOGIC_VECTOR(10 DOWNTO 0);
		 doutb : OUT STD_LOGIC_VECTOR(5 DOWNTO 0)
	  );
	END COMPONENT;

	-- position of read/write heads on drum
	-- Note that the read addresses are all one higher than expected, to compensate for 1 cycle latching  delay in the memory.

	subtype bitadr is std_logic_vector (10 downto 0);	-- bit address within a track
	
	constant readadr_ini:  bitadr := std_logic_vector (to_signed (  1, 11));	-- main memory and register read heads
	constant memwradr_ini: bitadr := std_logic_vector (to_signed (  0, 11));	-- main memory write address, readadr -1
	constant writeadr_ini: bitadr := std_logic_vector (to_signed ( 32, 11)); 	-- circulating register write heads
	constant staradr_ini:  bitadr := std_logic_vector (to_signed (-32, 11));	-- Astar read head: 65 bits downstream from write head
	
	signal writeadr: bitadr := writeadr_ini; 		-- main memory read heads; register write heads
	signal readadr:  bitadr := readadr_ini;		-- register read heads: 32 bits downstream
	signal staradr:  bitadr := staradr_ini;		-- Astar read head: 65 bits downstream
	signal memwradr: bitadr := memwradr_ini;		-- main memory write address, read adr -1
	signal muxadr: bitadr;								-- multiplexed address for main memory - either read or write, depending on W
	
	-- sync track patterns, S1..S3
	subtype 	sync_word is std_logic_vector (31 downto 0);
	signal 	s1_upstream: sync_word := x"80000000";
	signal 	s1_word: sync_word := x"80000000";
	constant	s2_word: sync_word := x"00003ffc";
	constant	s3_word: sync_word := x"800f3f00";
	
	-- internal intermediate output signals 
	-- (used to derive positive and inverted outputs).
	signal Aint, Rint, Cint, Vint, Asint: std_logic;
	signal S1int, S2int, S3int, S1up: std_logic;

begin

	-- memory blocks for re-circulating registers
	
	A_circulate: circulate 
	port map (
		clka 		=> clk,
		wea 		=> "1",
		addra		=> writeadr,
		dina(0)	=> Ap,
		clkb 		=> clk,
		addrb		=> readadr,
		doutb(0) => Aint
	);

	As_circulate: circulate 
	port map (
		clka 		=> clk,
		wea 		=> "1",
		addra		=> writeadr,
		dina(0)	=> Ap,
		clkb 		=> clk,
		addrb		=> staradr,
		doutb(0) => Asint
	);

	R_circulate: circulate 
	port map (
		clka 		=> clk,
		wea 		=> "1",
		addra		=> writeadr,
		dina(0)	=> Rp,
		clkb 		=> clk,
		addrb		=> readadr,
		doutb(0) => Rint
	);

	C_circulate: circulate 
	port map (
		clka 		=> clk,
		wea 		=> "1",
		addra		=> writeadr,
		dina(0)	=> Cp,
		clkb 		=> clk,
		addrb		=> readadr,
		doutb(0) => Cint
	);
	
	-- main memory instance
	inst_memory: memory
	PORT MAP (
		clka => clk,
		wea(0) => W,
		addra(10 downto 0) 	=> muxadr, 	-- bit address within track
		addra(16	downto 11)	=> P,			-- track address
		dina(0)					=> Vp,		-- data in (from logic module)
		douta(0)					=> Vint,		-- data out (to logic module)
	 
		 -- video display port - read only
		clkb => vid_clk,
		web(0) => '0',
		addrb => vid_adr,
		dinb(0) => '0',
		doutb(0) => vid_data
	);

	-- copy of registers and sync tracks for video output
	-- update this as soon as registers are written (32 bit upstream from read heads)
	reg_cache: cache
	  PORT MAP (
		 clka => clk,
		 wea => "1",
		 addra => writeadr,		-- upstream write head position
		 dina => S3int & S2int & S1up & Ap & Rp & Cp,
		 clkb => vid_clk,
		 addrb => vid_adr (10 downto 0),
		 doutb => vid_regs
	  );

--	-- copy of registers and sync tracks for video output
--	reg_cache: cache
--	  PORT MAP (
--		 clka => clk,
--		 wea => "1",
--		 addra => memwradr,		-- not readadr! one add'l tick delay due to buffering
--		 dina => S3int & S2int & S1int & Aint & Rint & Cint,
--		 clkb => vid_clk,
--		 addrb => vid_adr (10 downto 0),
--		 doutb => vid_regs
--	  );

	-- drive outputs and inverted outputs (asynchronous)
	A <= Aint;		An <= not Aint;
	R <= Rint; 		Rn <= not Rint;
	C <= Cint;		Cn <= not Cint;
	V <= Vint;		Vn <= not Vint;
	As <= Asint;	Asn <= not Asint;
	S1 <= S1int;	S1n <= not S1int;
	S2 <= S2int;	S2n <= not S2int;
	S3 <= S3int;	S3n <= not S3int;		
	
	drumpos <= readadr;

	-- multiplexed read/write address for main memory
	muxadr <= memwradr when (W = '1') else  readadr;
	
	-- read/write operations to "drum" (synchronous)
	process (clk, sec_reset) is
	
		-- current sector number (for S1 sync track)
		-- Start with Sector 0 for physical RAM address 0.
		-- Note that S1 denotes the number of the *following* sector! 
		-- sector count informs to the s1_upstream value, which is  further word ahead.
		-- Hence give it a "headstart" of two sectors, at 50 (= 64-7-7), which follows sector 0.
--		constant sector_ini: std_logic_vector (5 downto 0) := "111001";
		constant sector_ini: std_logic_vector (5 downto 0) := "110010";
		variable sector: std_logic_vector (5 downto 0) := sector_ini;
		
	begin
		if sec_reset = '0' then
			if rising_edge (clk) then
			
				-- output sync tracks
				S1up	<= s1_upstream (to_integer (unsigned (readadr (4 downto 0))));
				S1int	<= s1_word (to_integer (unsigned (readadr (4 downto 0)))); 
				S2int	<= s2_word (to_integer (unsigned (readadr (4 downto 0)))); 
				S3int	<= s3_word (to_integer (unsigned (readadr (4 downto 0)))); 
					
				-- monitor end of sector
				if readadr (4 downto 0) = "11111" then 	-- end of sector

					-- about to reach next word on drum; prepare next sector number
					sector := sector - 7;	-- sectors are interleaved, not consecutive 
					s1_word <= s1_upstream;
					s1_upstream <= "10000" & sector & "0000000000000" & sector & "00";

				end if;
				
				-- "rotate drum": increment address counters within track
				writeadr <= writeadr + 1;
				readadr <= readadr + 1;
				staradr <= staradr + 1;
				memwradr <= memwradr + 1;
						
			end if;
		else
			-- async. reset of rotary position (for xmodem transfers)
			writeadr <= writeadr_ini; 		-- main memory; register write heads
			readadr <= readadr_ini;			-- register read heads: 32 bits downstream
			staradr <= staradr_ini;			-- Astar read head: 65 bits downstream
			memwradr <= memwradr_ini;		-- main memory write address, read adr -1
			sector := sector_ini;
		end if;
	end process;

end Behavioral;

