----------------------------------------------------------------------------------

-- LGP-30 drum computer simulation
-- Flexowriter simulation (via serial terminal)
--
-- jm 1.8.2016
--
-- 1.11.2016 ASCII encoding/decoding changed - 1/l = letter l, Pi, Sigma, Delta changed

-- Required functionality:
--
-- Operates at 19200Bd 8n1, but limits input speed to 200cps and print speed to 10cps via delays,
-- to ensure proper handshake with the logic unit. Needs hardware handshake on UART input side (CTS), 
-- to enable reception of text files as a punch-tape simulation (wait until next P00 command sets Tread).
--
-- *** Output:
-- Triggered by <Ttrig> rising edge.
-- Keep track of Uppler/Lower Shift (control characters).
-- Convert all other characters to ASCII, with code translation depending on current shift setting.
-- Send via UART at fixed baud rate. 
-- Send a start signal (<bs> ???) to the CPU when printing is complete.
--
-- *** Input:
-- Prepared from LGP-30 CPU by triggering <Tread>.
-- This starts receipt of a "burst" of characters from the Flexowriter.
-- In "manual input" mode (panel switch), always allow input and "force feed" to logic unit??? (Documentation not clear here.)
-- The LGP-30 CU must always be fast enough to digest subsequent characters;
-- limit the input speed by selecting a slow baud rate!
--
-- Upon receipt of a "Cond Stop" code (encoded as ESC on the ASCII side), set <bs> Signal to CPU.
-- (This signal must be sent at least after every 8th 4-bit character, to avoid A overflow.
-- It is the responsibilty of the operator to send COnd Stop or prepare the punched tape accordingly.)
-- Stop transmission to the LGP-30 until a further <Tread> pulse is received.
--
-- According to the LGP-30 Operations Manual, several control characters are NOT forwarded to the LGP-30 on input:
-- Upper & Lower case shift, color shift, CR, backspace, delete.
-- TAB is an exception here and gets forwarded (as "00" -- different from Tx code!).
-- No need to keep track of upper/lower case state on input; 
-- the LGP-30 does not distinguish between these and just receives the 6-bit (or 4-bit) codes.
-- 
-- <Tstart> signal gets reset to zero via an internal timer.
-- (Simulate the cam with fixed timing used in the original Flexowriter.)
-- Same for Ps signal.
--
-- *** Data format conventions
-- The 6bit inputs and outputs (Tp, PPs) follow the LGP-30 convention: Bit 1 = MSB, bit 6 = LSB!
-- By declaring them as std_logic_vectors (1 to 6), VHDL will follow this same convention,
-- and the decimal representations used in the code conversion lookup tables will work.
--
-- Special character translations differ from Stuttgart tape convention:
-- ^ (94) = Delta, ~ (126) = Pi, & (38) = Sigma
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library work;
use work.constants.all;


entity flexowriter is
	port (
	
		-- inside simulated LGP-30
		lgp_clk: in std_logic;							-- LGP-30 bit clock; determines timing toards LGP-30 logic unit
		Tp: in std_logic_vector (1 to 6);			-- character to be printed/punched
		PPs, PPsn: out std_logic_vector (1 to 6);	-- character from keyboard or tape reader
		Ps, Psn: out std_logic;							-- signal that one char was received
		Ttrig: in std_logic;								-- trigger printing
		Tread: in std_logic;								-- trigger reading of next 4/6 bit input
		Tdelay: in std_logic;							-- trigger delay in Flexwriter (ignored in simulation)
		Tstart: out std_logic;							-- received Cond Stop code; restart logic unit from blocked state
		
		-- connection to UART unit
		io_clk: in std_logic;							-- I/O clock
		uart_data_tx: out std_logic_vector(7 downto 0);		
		uart_data_rx: in std_logic_vector(7 downto 0);		
		write_to_uart_tx: out std_logic;				-- trigger 1 byte write to Tx buffer
		read_from_uart_rx: out std_logic;			-- trigger 1 byte read from Rx buffer
		uart_rx_half_full: in std_logic; 			-- Rx buffer >= half full?
		uart_rx_data_present: in std_logic; 		-- Rx buffer has character?

		-- enable output of ESC sequences
		use_ESC: in std_logic;
		
		-- read start/stop control: connections to control panel
		readActive: out std_logic;						-- signals that FlexoWriter is ready to read (for read indicator led)
		readToggle: in std_logic;						-- alternate between "start read" and "stop read"; actual status controlled by Flexowriter
		
		-- handshake with xmodem unit
		xmodem_active: buffer std_logic;				-- xmodem transfer is active
		xmodem_done: in std_logic;						-- strobes high when xmodem is done (or has failed)
		
		-- simulate printing noise (via external relay)
		noise: out std_logic;
		
		-- debug outputs
		test: out std_logic_vector (7 downto 0)
	);
end flexowriter;

architecture Behavioral of flexowriter is

--
-- Signals to manage UART/LGP-30 interaction
--

-- state flags
signal Ttrig_prev: std_logic := '0';				-- previous clock's state of Ttrig, for rising edge detection	
signal readTogglePrev: std_logic := '0';			-- previous clock's state of readToggle, for edge detection
signal state_upper: natural range 0 to 1 := 0;	-- upper case output
signal state_input: std_logic := '1';				-- input is active (disabled by cond'l stop character)
signal state_red: std_logic := '0';					-- color shift state of connected terminal

signal write_strobe: std_logic := '0';				-- cache write_to_uart_tx (need to read it below)

-----------------------------------------------------------------------------------------
-- timers
-- durations given in lgp_clk cycles (120 kHz in regular LGP-30),
-- to enable operation with variable CPU clock speed
-----------------------------------------------------------------------------------------

-- Ps pulse length (one character received, CPU to shift into accumulator)
constant Ps_timeout_init: natural := 125;			-- pulse length 1 ms @125 kHz CPU clock
signal Ps_timeout: natural range 0 to Ps_timeout_init := 0;	

-- Tstart pulse length (cond'l stop code received, CPU to resume processing)
constant Tstart_timeout_init: natural := 5*125;	-- pulse length 5ms @125 kHz CPU clock
signal Tstart_timeout: natural range 0 to Tstart_timeout_init := 0;	

-- Limit the rate of input characters fed to LGP-30 from UART.
--
-- Timer in positive range indicates that the LGP-30 is not ready for input yet.
-- Timer runs down into the negative range for slow (manual) input;
-- this is used to distinguish manual input and henerate relay clicks only then (and not for "tape" reading)
constant Input_timeout_init: integer := 24*125;		-- delay length 24ms (166 cps) @125 kHz CPU clock 
																	-- increased from 6 ms 7.10.16.  12ms too fast, 24ms OK. 
constant Input_timeout_min: integer := -10*125;		-- if timer runs down this far, we are dealing with manual input
signal Input_timeout: integer range Input_timeout_min to Input_timeout_init := 0;	

-- limit rate of output characters fed from LGP-30 to UART
constant Print_timeout_init: natural := 100*125;	-- delay length 100ms (10 cps) @125 kHz CPU clock
																	-- increased from 90ms 30.10.16.  93ms OK for Mondlandung, 90ms too short. 
constant Print_timeout_click: natural := Print_timeout_init/2;		-- clicking relay is on while timer > this
signal Print_timeout: natural range 0 to Print_timeout_init := 0;	

-- Relay click duration for echo'ed input characters
constant click_timer_start: natural := 40*125;
signal click_timer: natural range 0 to click_timer_start := 0;

-----------------------------------------------------------------------------------------
-- character constants and translation
-----------------------------------------------------------------------------------------

-- character constants - LGP-30 6-bit code
constant char_upper: std_logic_vector (1 to 6) := "001000";	-- shift to uppercase
constant char_lower: std_logic_vector (1 to 6) := "000100";	-- shift to lowercase
constant char_color:	std_logic_vector (1 to 6) := "001100";	-- color shift (toggle red/black)
constant char_tab:	std_logic_vector (1 to 6) := "000000";	-- tab (when read from Flexowriter and sent to CPU)
constant char_inval: std_logic_vector (1 to 6) := "111111";	-- invalid character (internal use only)

-- character constants - ASCII code
constant ascii_esc: std_logic_vector (7 downto 0) := "00011011";
constant ascii_tab: std_logic_vector (7 downto 0) := "00001001";
constant ascii_prime: std_logic_vector (7 downto 0) := "00100111";
constant ascii_spc: std_logic_vector (7 downto 0) := "00010000";

-- lookup table - Flexowriter code to ascii
type ascii_table is array (0 to 63) of natural range 0 to 255;	
type ascii_lookup is array (0 to 1) of ascii_table;
constant ascii: ascii_lookup := 
(
	( 	-- lower case
		0, 122, 48, 32, 0, 98, 108, 45, 62, 121,
		50, 43, 0, 114, 51, 59, 13, 105, 52, 47,
		8, 100, 53, 46, 9, 110, 54, 44, 0, 109,
		55, 118, 39, 112, 56, 111, 0, 101, 57, 120,
		0, 117, 102, 0, 0, 116, 103, 0, 0, 104,
		106, 0, 0, 99, 107, 0, 0, 97, 113, 0,
		0, 115, 119, 7 
	),
	( 	-- upper case
		0, 90, 41, 32, 0, 66, 76, 95, 60, 89,
      42, 61, 0, 82, 34, 58, 13, 73, 94, 63,
		8, 68, 37, 93, 9, 78, 36, 91, 0, 77,
		126, 86, 39, 80, 38, 79, 0, 69, 40, 88,
		0, 85, 70, 0, 0, 84, 71, 0, 0, 72,
		74, 0, 0, 67, 75, 0, 0, 65, 81, 0,
		0, 83, 87, 7 
	)
);		  
--(
--	(0,122,48,32,0,98,108,45,0,121,50,43,0,114,51,59,13,105,52,47,8,100,53,46,9,110,54,44,0,109,55,118,27,112,56,111,0,101,57,120,0,117,102,0,0,116,103,0,0,104,106,0,0,99,107,0,0,97,113,0,0,115,119,127),
--	(0,90,41,32,0,66,76,95,0,89,42,61,0,82,34,58,13,73,60,63,8,68,37,93,9,78,36,91,0,77,35,86,27,80,62,79,0,69,40,88,0,85,70,0,0,84,71,0,0,72,74,0,0,67,75,0,0,65,81,0,0,83,87,127)
--);

-- lookup table - ascii to Flexowriter.
-- starts with $20 (ascii space)
type flex_lookup is array (32 to 127) of natural range 0 to 63;
constant flex: flex_lookup :=
( 	
	-- digits and punctuation
	3, 3, 14, 3, 18, 22, 34, 32,
	38, 02, 10, 11, 27, 07, 23, 19, 2, 6,
	10, 14, 18, 22, 26, 30, 34, 38, 15, 15,
	8, 11, 4, 19, 
	
	-- capital letters
	3, 57, 5, 53, 21, 37,
	42, 46, 49, 17, 50, 54, 6, 29, 25, 35,
	33, 58, 13, 61, 45, 41, 31, 62, 39, 9,
	1, 27, 3, 23, 18, 7,
	
	-- lowercase letters
	3, 57, 5, 53, 21, 37,
	42, 46, 49, 17, 50, 54, 6, 29, 25, 35,
	33, 58, 13, 61, 45, 41, 31, 62, 39, 9,
	1, 27, 3, 23, 30, 63
);
--(
--	3,0,14,30,26,22,0,0,38,2,10,11,27,7,23,19,2,6,10,14,18,22,26,30,34,38,15,15,18,11,34,19,0,57,5,53,21,37,42,46,49,17,50,54,6,29,25,35,33,58,13,61,45,41,31,62,39,9,1,27,0,23,0,7,0,57,5,53,21,37,42,46,49,17,50,54,6,29,25,35,33,58,13,61,45,41,31,62,39,9,1,0,0,0,0,63
--);

-- Escape sequences

signal esc_count: natural range 0 to 4 := 0;
signal esc_index: natural range 0 to 1 := 0;

--constant ESC: character := character'val (27);
--type esc_seq is array (0 to 1, 0 to 3) of character;
--constant esc_table: esc_seq := (
--	(ESC, '[', '0', 'm'),
--	(ESC, '[', '1', 'm')
--);

type esc_seq is array (0 to 1, 0 to 3) of std_logic_vector (7 downto 0);
constant esc_table: esc_seq := (
	(ascii_esc, "01011011", "00110000", "01101101"),
	(ascii_esc, "01011011", "00110001", "01101101")
);
	

begin
 
  -----------------------------------------------------------------------------------------
  -- I/O from & to LGP-30
  -----------------------------------------------------------------------------------------
  
	rxtx: process (io_clk)
	variable rxchar: std_logic_vector (5 downto 0) := "000000";
	begin
		if rising_edge (io_clk) then
		
			if xmodem_active = '1'then
				-- XModem unit has control.
				-- Only possible action is to end xmodem mode upon receipt of "done" strobe:
				if xmodem_done = '1' then xmodem_active <= '0'; end if;	
				
			else
				-- Flexowriter operation, not XModem
				
				-- signal to the control panel whether input is active or blocked 
				readActive <= (not uart_rx_half_full) and state_input;
				
				-- toggle input state when control panel "start/stop read" switch signals manual change
				if readToggle /= readTogglePrev  then	-- when state changes
					state_input <= not state_input;
				end if;
				readTogglePrev <= readToggle;

				------ count down timers -------
				
				-- Timers manage pulse lengths and delay times towards the LGP-30 logic unit.
				-- This is driven by the logic unit's clock, so the timing is correct for slowed-down logic clocks.
				-- Note that the logic unit clock is asymmetrical -- high for one I/O clock cycle only.
				
				if (lgp_clk = '1') then
				
					-- Pulse duration for "Start from blocked state" (bs, to logic unit)
					if Tstart_timeout > 0 then 
						Tstart_timeout <= Tstart_timeout -1; 
					else
						Tstart <= '0';
					end if;
					
					-- Pulse duration for P* and P*1..P*6 (to logic unit)
					if Ps_timeout > 0 then 
						Ps_timeout <= Ps_timeout -1; 
					else
						Ps <= '0'; Psn <= '1';
						PPs <= "000000"; PPsn <= "000000";	-- switch back to passive state: neither set nor reset P flip-flops
					end if;

					-- limit input rate: delay next character to be sent
					-- (otherwise the UART buffer might overrun the logic unit)
					if Input_timeout > Input_timeout_min then 
						Input_timeout <= Input_timeout -1; 
					end if;

					-- limit print rate: delay start signal confirming completion of printing a character.
					-- (logic unit might be busy with other code before entering stop state)
					if Print_timeout > 0 then 
						Print_timeout <= Print_timeout -1;
					end if;
					if Print_timeout = 1 then
						-- always signal "printing complete" after timeout, since serial Tx is much faster
						Tstart <= '1';											-- signal that printing is complete
						Tstart_timeout <= Tstart_timeout_init;			-- start timer for Tstart pulse length
					end if;
					
					-- click duration for echo'ed characters
					if click_timer > 0 then click_timer <= click_timer-1;
					end if;

				end if;
				
				-- Reset strobe signals
				-- This can be overwritten by setting the strobes further down
				--
				write_strobe <= '0';
				read_from_uart_rx <= '0';

				------- ESC sequence ongoing? -------

				if (esc_count>0) and (write_strobe = '0') then
					uart_data_tx <= esc_table (esc_index, 4-esc_count);
					esc_count <= esc_count-1;
					write_strobe <= '1';
	
				------- output from LGP-30 -------

				elsif (Ttrig = '1') and (Ttrig_prev = '0') then		-- have a character to print
					
					if tp = char_upper then
						-- keep track of shift state.
						-- These special characters do not get "printed" via the UART,
						-- but will be confirmed via a Start signal to the CPU anyway.
						state_upper <= 1;
						Print_timeout <= Print_timeout_init;		-- Tstart pulse will follow after a simulated printing delay
					elsif tp = char_lower then
						state_upper <= 0;
						Print_timeout <= Print_timeout_init;		-- Tstart pulse will follow after a simulated printing delay
					elsif (tp = char_color) and (use_ESC = '1') then
						-- toggle color shift state
						if (state_red = '0') then 
							esc_index <= 1;	-- set to red
							state_red <= '1';
						else 
							esc_index <= 0;
							state_red <= '0';
						end if;
						esc_count <= 4;
						state_red <= NOT state_red;
						Print_timeout <= Print_timeout_init;		-- Tstart pulse will follow after a simulated printing delay							
					else
						-- code conversion ...
						if std_logic_vector (to_signed (ascii (state_upper)(to_integer (unsigned (tp))), 8)) /= 0 then
							uart_data_tx <= std_logic_vector (to_signed (ascii (state_upper)(to_integer (unsigned (tp))), 8));
							write_strobe <= '1';					-- trigger transmission if valid character
						end if;
						Print_timeout <= Print_timeout_init;		-- Tstart pulse will follow after a simulated printing delay
					end if;
					
				end if;
				Ttrig_prev <= Ttrig;
				
				------- input to LGP-30 -------

				if (Tread = '1') then 
					state_input <= '1';									-- LGP-30 is asking for input
					Input_timeout <= Input_timeout_init;			-- start new timer (do not overrun LGP-30 right after read trigger)				
				end if;

				-- Check for initiation of XModem upload or download.
				-- Hand over control to XModem immediately;
				-- XModem unit must handle the starting character in the UART buffer
				if (uart_rx_data_present = '1') and ((uart_data_rx = xmodem_start_rx) or (uart_data_rx = xmodem_start_tx)) then
					xmodem_active <= '1';
		
				-- Input a character if requested from CPU.
				-- Act only if CPU requests input, data are present, and overrun protection delay is over.
				-- ESC is an exception, since the "start" key can be depressed at any time.
				
				-- Note: While ESC is detected even when the CPU is not expecting input,
				-- it cannot "bypass" other characters which may be in the input buffer, and is not detected in that case.
				-- This is probably different from the real LGP-30, where the Start key used dedicated wiring.
				
				elsif (uart_rx_data_present = '1') 
					and ((state_input = '1' and Input_timeout <= 0) or uart_data_rx = ascii_esc) then

					-- LGP is ready to receive data, and UART has data
					rxchar := char_inval;								-- start value - not a valid input character
					
					if (Input_timeout = Input_timeout_min) then
						Click_timer <= Click_timer_start;
					end if;
					
					-- Echo back to terminal.
					-- This should reduce confusion when using PC UARTs with heavy buffering (USB/UART adapters),
					-- by showing on the PC terminal which characters have actually been digested by the LGP-30.
					-- Do not echo ESC (cond'l stop key), as the terminal would expect a control sequence after that.

					if uart_data_rx /= ascii_esc then
						uart_data_tx <= uart_data_rx;
						write_strobe <= '1';							-- trigger transmission
					end if;

					-- ' decoded as conditional stop, signals that input word is complete.
					-- Trigger Tstart, clear state_input ("stop tape").

					if (uart_data_rx = ascii_prime) then
						state_input <= '0';								-- stop input flow
						Tstart <= '1';										-- signal that input is complete
						Tstart_timeout <= Tstart_timeout_init;		-- start timer for Tstart pulse length

					-- ESC decoded as start key, signals that input word is complete.
					-- Trigger Tstart, but do not clear state_input. 
					-- (No need to stop "tape". Allow future ESC to be read for re-starting the CPU.)

					elsif (uart_data_rx = ascii_esc) then
						Tstart <= '1';										-- signal that input is complete
						Tstart_timeout <= Tstart_timeout_init;		-- start timer for Tstart pulse length

					-- filter out control characters;
					-- do code conversion 

					elsif (uart_data_rx = ascii_tab) then			-- the only control character we need to convert
						rxchar := char_tab;
					elsif (uart_data_rx >= ascii_spc) then
						rxchar := std_logic_vector (to_unsigned (flex (to_integer (signed (uart_data_rx))), 6));
					end if;
					
					-- transfer to LGP-30 
					
					read_from_uart_rx <= '1';							-- signal to UART that we have read that char
					Input_timeout <= Input_timeout_init;			-- start new timer (limit char rate to LGP-30)
																				-- do this also for characters not sent to the logic unit,
																				-- to avoid duplicate processing in this input routine
					if rxchar /= char_inval then						-- have something valid to send to LGP-30 
						PPs <= rxchar;										-- provide non-inverted and inverted data
						PPsn <= not rxchar;
						Ps <= '1'; Psn <= '0';							-- signal char to logic unit
						Ps_timeout <= Ps_timeout_init;				-- timer will end Ps pulse
					end if;
				end if;	-- state_input='1'...
			end if;	-- xmodem_active

			-- operate clicking relay 
			
			if (Print_timeout > Print_timeout_click) or (Click_timer > 0) then noise <= '1';
			else noise <= '0';
			end if ;
	
		end if;	-- rising_edge (io_clk)
	end process rxtx;

	write_to_uart_tx <= write_strobe;
	
------------------------------------------------
-- assign test outputs for debugging
------------------------------------------------

	test <= (others => '0');

end Behavioral;

