----------------------------------------------------------------------------------

-- LGP-30 drum computer simulation
-- HDMI output (magnetic drum content, possibly further status info in the future)
--
-- jm 24.8.2016	original version
-- jm  2.8.2017 	shift_target is written/read in different clock domains. 
--						Introduced an additional buffer stage to avoid timing warnings.
-- jm 22.10.2017	flip-flop display and decoding added; drum positioning with wrap-around

----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
-- use IEEE.STD_LOGIC_ARITH.ALL;

library work;
use work.constants.all;


entity hdmi is

	port (
		-- system clock
		clk100: in std_logic;
	
		-- data input (main drum memory)
		clk: out std_logic;
		adr: buffer std_logic_vector (16 downto 0);
		data: in std_logic;
		regs: in std_logic_vector (5 downto 0);		-- S3, S2, S1, A, R, C
		
		-- data for display of r/w heads (used at slow clock speeds only)
		drumpos: in std_logic_vector (10 downto 0);	-- current drum rotary position
		track: in std_logic_vector (5 downto 0);		-- current LGP-30 track address
		write_en: in std_logic;								-- write enable for LGP-30 heads

		-- flip-flops for status display
		F, G, H, K, L: in std_logic;
		P: in std_logic_vector (1 to 6);
		Q: in std_logic_vector (1 to 4);
			
		-- display mode control
		io_clk: in std_logic;
		right: in std_logic;				-- pulses high to move display right (encoder controlled)
		left: in std_logic;				-- pulses high to move display left (encoder controlled)
		animate: in std_logic;			-- animate the display (at slow clock speeds)
		showff: in std_logic;			-- show flip-flops
		
		-- HDMI output
		hdmi_p : out std_logic_vector(3 downto 0);
		hdmi_n : out std_logic_vector(3 downto 0);
		
		-- debug outputs
		test: out std_logic_vector (7 downto 0)
	);
end hdmi;

architecture Behavioral of hdmi is

   COMPONENT xga_clocks
		PORT(
			clk100			 : IN std_logic;
			pixel_clock     : OUT std_logic;
			clock_5x     	 : OUT std_logic
      );
   END COMPONENT;

   COMPONENT hdmi_encoder
		PORT(

			-- pixel and TMDS clock
			pixel_clock     : IN std_logic;
			clock_5x     	 : IN std_logic;
		
			-- input: XGA signal 
			red, green, blue: IN std_logic_vector(7 downto 0);
			blank           : IN std_logic;
			hsync           : IN std_logic;
			vsync           : IN std_logic;
			
			-- HDMI output
			hdmi_p : OUT std_logic_vector(3 downto 0);
			hdmi_n : OUT std_logic_vector(3 downto 0)

     );
   END COMPONENT;


	-- XGA timing constants
	constant h_rez        : natural := 1024;
   constant h_sync_start : natural := 1024+24;
   constant h_sync_end   : natural := 1024+24+136;
   constant h_max        : natural := 1343;
   constant v_rez        : natural := 768;
   constant v_sync_start : natural := 768+3;
   constant v_sync_end   : natural := 768+3+6;
   constant v_max        : natural := 805;
	
	-- Color levels for bit display and track/sector digits
	constant COL_OFF: 		std_logic_vector (7 downto 0) := "00000000";
	constant COL_WHITE: 		std_logic_vector (7 downto 0) := "11111111";
	constant COL_GREY:		std_logic_vector (7 downto 0) := "01000000";
	constant COL_DIGIT:		std_logic_vector (7 downto 0) := "11111111";

	-- Color levels for word structure overlay
	constant BLUE_OFF: 		std_logic_vector (7 downto 0) := "00000000";
	constant BLUE_TRACK: 	std_logic_vector (7 downto 0) := "11111111";
	constant BLUE_SECTOR: 	std_logic_vector (7 downto 0) := "11000000";
	constant BLUE_OPCODE: 	std_logic_vector (7 downto 0) := "11000000";
	constant BLUE_OTHER: 	std_logic_vector (7 downto 0) := "10000000";
	
	-- Color levels for R/W head
	constant COL_PASSIVE: 	std_logic_vector (7 downto 0) := "10100000";
	constant COL_READ: 		std_logic_vector (7 downto 0) := "11111111";
	constant COL_WRITE: 		std_logic_vector (7 downto 0) := "11111111";
	
	-- color levels for Flip-Flops
	constant RED_ON: 			std_logic_vector (7 downto 0) := "11000000";
	constant RED_OFF: 		std_logic_vector (7 downto 0) := "01100000";
	constant GREEN_ON: 		std_logic_vector (7 downto 0) := "10100000";
	constant GREEN_OFF: 		std_logic_vector (7 downto 0) := "00100000";
	
	-- scale of bits on the screen
	constant subline_max: natural := 9;									-- height of a track -1
	constant subcol_max: natural := 2;									-- width of a bit -1
	
	-- vertical screen division - absolute screen lines
	-- Vertical position of the drum display will change slightly if flip-flop display is enabled.
	-- Actual settings in the code below.
	signal 	V_SEC_MIN: 		natural range 0 to v_rez-1;			-- sector numbers
	signal 	V_SEC_MAX: 		natural range 0 to v_rez-1;			-- letters are 8 pixels high
	signal	V_BITS_MIN: 	natural range 0 to v_rez-1;			-- track data begins. Must be multiple of subline period.
	signal   V_BITS_MAX: 	natural range 0 to v_rez-1;			-- 64 tracks, each 10 pixels high
	signal   V_REGS_MIN: 	natural range 0 to v_rez-1;			-- track data begins. Must be multiple of subline period.
	signal   V_REGS_MID:		natural range 0 to v_rez-1;			-- 3 tracks with circulating registers
	signal   V_REGS_MAX: 	natural range 0 to v_rez-1;			-- 6 tracks, each 10 pixels high

	constant FF_SIZE:			natural := 18;								-- size of square flip-flop "lamp"
	constant V_FF_MIN:		natural := 740;							-- flip-flop "lamp" display area
	constant V_FF_MAX:		natural := V_FF_MIN+FF_SIZE-1;
	constant V_FFLBL_MIN:	natural := V_FF_MIN+(FF_SIZE-8)/2;	-- flip-flop label text, 8 pixels high, centered in square
	constant V_FFLBL_MAX:	natural := V_FFLBL_MIN+8-1;
	
	-- horizontal position of sector numbers - pixels counted left to right, starting at bit 17 of each word
	constant H_SEC_HIGH:		natural := 1;								-- left edge of high sector no. digit
	constant H_SEC_LOW:		natural := H_SEC_HIGH+6;				-- left edge of low sector no. digit
	
	-- vertical position of track numbers - relative to subline counter
	constant V_TRK_MIN:		natural := 2;								-- track numbers
	constant V_TRK_MAX:		natural := V_TRK_MIN+7;					-- letters are 8 pixels high

	-- horizontal position of track numbers - from left edge of screen
	constant H_TRK_HIGH:		natural := 8;								-- left edge of high track no. digit
	constant H_TRK_LOW:		natural := H_TRK_HIGH+6;				-- left edge of low track no. digit
	constant H_TRK2_HIGH:	natural := h_rez-20;						-- left edge of high track no. digit
	constant H_TRK2_LOW:		natural := H_TRK2_HIGH+6;				-- left edge of low track no. digit
	
	-- horizontal position for drum display area - from left edge of screen
	constant H_BITS_LOW:		natural := 32;
	constant H_BITS_HIGH:	natural := h_rez-32;
	
	-- horizontal position of R/W head in animated mode - from left edge of screen
	constant HEADPOS:			natural := 6*32;							-- head position, in bits, from left edge of display area
	constant WRHEADPOS:		natural := HEADPOS-32;					-- write head for recirculating registers: 1 word upstream
	constant H_HEAD_LOW:		natural := H_BITS_LOW+(subcol_max+1)*HEADPOS;
	constant H_HEAD_HIGH:	natural := H_HEAD_LOW+3;
	constant H_WRHEAD_LOW:	natural := H_BITS_LOW+(subcol_max+1)*WRHEADPOS;
	constant H_WRHEAD_HIGH:	natural := H_WRHEAD_LOW+3;
	
	-- labels for register and sync tracks
	type track_text is array (0 to 1) of character;
	type track_table is array (0 to 5) of track_text;
	constant track_label: track_table := (
		(' ', 'C'),
		(' ', 'R'),
		(' ', 'A'),
		('S', '1'),
		('S', '2'),
		('S', '3')
	);
	
	-- positions and labels for flip-flop display
	type ff_label_t is array (0 to 14) of character;
	constant ff_label: ff_label_t := (
		'F', 'G', 'H', 'K', 'L', '1', '2', '3', '4', '1', '2', '3', '4', '5', '6'
	);	

	-- horizontal positions for flip-flop legend (pixels)
	type ff_pos_t is array (0 to 14) of natural range 0 to h_rez-1;
	constant ff_pos: ff_pos_t := (
		125, 150, 175, 325, 350, 475, 500, 525, 550, 725, 750, 775, 800, 825, 850
	);
	
	constant FF_POS_PHASE: natural := 200;
	constant FF_POS_OPCODE: natural := 575;
	constant FF_POS_PREG: natural := 875;
	
	signal ff_state: std_logic_vector (0 to 14);						-- cached copy of flip-flop signals (stable during full frame)
	signal ff_index: natural range 0 to 15;							-- index counts up flip-flops for each output line
	
	-- text for flip-flop legends
	constant NSTRINGS: natural := 28;
	constant nul: character :=  character'val(0);
	type ff_string_t is array (1 to NSTRINGS) of String (1 to 11);
	constant ff_strings: ff_string_t := (
										-- FGH
		"Wait Instr" & nul,		-- 000
		"Phase 5   " & nul,		-- 001
		"Read Instr" & nul,		-- 010
		"Phase 8   " & nul,		-- 011
		"Wait Data " & nul,		-- 100
		"Phase 6   " & nul,		-- 101
		"Exec Instr" & nul,		-- 110
		"Phase 7   " & nul,		-- 111
		"Blocked   " & nul,		-- 000 and Q2=0
		
		"Z Stop    " & nul,
		"Y StoreAdr" & nul,
		"B Bring   " & nul,
		"R StoreRet" & nul,
		"I Input   " & nul,
		"D Div     " & nul,
		"N Multiply" & nul,
		"M Multiply" & nul,
		"P Print   " & nul,
		"E Extract " & nul,
		"U Uncond  " & nul,
		"T Test    " & nul,
		"H Hold    " & nul,
		"C Clear   " & nul,
		"A Add     " & nul,
		"S Subtract" & nul,
		
		"Track #*  " & nul,	-- #* will be replaced by P value	
		"Print #*  " & nul,
		"Input #*  " & nul
	);
	signal ff_legend: natural range 0 to NSTRINGS;			-- legend being currently printed (0=idle)
		
	-- XGA (output) counters
	signal h_count: natural range 0 to 2047 := 0;
	signal v_count: natural range 0 to 1023 := 0;
	signal subline: natural range 0 to subline_max := 0;			-- every track takes 11 screen lines
	signal subcol: natural range 0 to subcol_max := 0;				-- every pixel takes 3 screen columns
	signal sec_col: natural range 0 to 32*(subcol_max+1) := 0;	-- count columns during sector number printing
	signal red, green, blue: std_logic_vector(7 downto 0);		-- current pixel intensity
	
	signal pixel_clk, clk_5x: std_logic;
	signal vsync, hsync, blank: std_logic;
	signal sector: natural range 0 to 63 := 0;						-- count interleaved sector numbers
	
	signal t_index, s_index, f_index, l_index: natural range 0 to 6*128-1 := 0; -- index to character table
	signal l_char: natural range 1 to 11;								-- character counter inside strings
	signal l_col: natural range 0 to 6;									-- column counter inside characters
	
	-- display shift (left/right)
	signal shift, shift_target, shift_target_unsync: integer range -1024 to 1023 := 0;	-- display shifted by X bits
	signal animate_prior: std_logic := '0';							-- prior state of animate input, to manage smooth transitions

begin

	inst_xga_clocks: xga_clocks 
		port map (
			clk100 => clk100,
			pixel_clock => pixel_clk,
			clock_5x => clk_5x
		);
	
	inst_hdmi_encoder: hdmi_encoder
		port map (
			pixel_clock => pixel_clk,
			clock_5x => clk_5x,
			
			red => red, 
			green => green,
			blue => blue,
			blank => blank,
			vsync => vsync,
			hsync => hsync,
		
			hdmi_p => hdmi_p,
			hdmi_n => hdmi_n
		);
		
	-- output pixel clock (for memory access)
	clk <= pixel_clk;
	
	-- monitor encoder (shift display left/right)
	process (io_clk) is
	begin
		if rising_edge (io_clk) then
			if (animate='0') and (animate_prior='1') then
				-- switching back to stationary mode:
				-- start with the current drum position, rounded to word boundary
				shift_target_unsync <= (2048 + 32 - HEADPOS - to_integer (unsigned (drumpos)))/32;	
			else
				-- in stationary mode, step by whole words 
				if (right='1') then shift_target_unsync <= (shift_target_unsync-1);
				elsif (left='1') then shift_target_unsync <= (shift_target_unsync+1);
				end if;
			end if;
			animate_prior <= animate;
		end if;
	end process;
	
	process (pixel_clk) is
	
	variable byte: std_logic_vector (7 downto 0);
	variable bit: Std_logic;
	variable delta: integer range -1024 to 1023;						-- calculate shift directions with wrap-around (LGP-30 bits along track)
	
   begin
      if rising_edge(pixel_clk) then
		
			-- copy of shift_target sync'd with the pixel clock
			shift_target <= shift_target_unsync * 32;
		
         -- set intensity output and blank signals
			if h_count < h_rez and v_count < v_rez then
			
				-- we are within the visible screen area.
				blank <= '0';

				-- clear pixel output first; will be overwritten by subsequent code.
				-- all following output intensities are additive.
				red <= COL_OFF; green <= COL_OFF; blue <= COL_OFF;
				
				-- update drum addresses (bit and sector)
				if h_count = h_rez-1 then
					-- new line in drum display area
					-- update drum memory address
					adr (10 downto 0) <= std_logic_vector (to_unsigned(2047-shift, 11)); -- go to end of track (displaying high bit first) 
					subcol <= 0;
				elsif h_count >= H_BITS_LOW then
					-- move to next bit within track if required.
					-- scanning the display left to right => show high bit first.
					-- no overflow into track address, please!
					if subcol = subcol_max then
						adr(10 downto 0) <= std_logic_vector (unsigned(adr(10 downto 0))-1);
					end if;
					subcol <= (subcol+1) mod (subcol_max+1);		
				end if;

				if (v_count <= V_SEC_MAX) and (v_count >= V_SEC_MIN) and (h_count >= H_BITS_LOW) and (h_count <= H_BITS_HIGH) then
					
					-- show sector numbers
					sector <= (512 - 7*to_integer (unsigned (adr (10 downto 5)))) mod 64;	-- count interleaved sectors
					if to_integer (unsigned (adr (4 downto 0))) = 17 then
						-- new word begins (backwards)
						sec_col <= 0;		-- begin counting screen columns
					elsif to_integer (unsigned (adr (4 downto 0))) < 17 then
						case sec_col is
							when H_SEC_HIGH-1 =>		
								s_index <= char_zero+6*(sector/10)+1;						

							when H_SEC_LOW-1 	=>		
								s_index <= char_zero+6*(sector MOD 10)+1;

							when H_SEC_HIGH to H_SEC_HIGH+4 | H_SEC_LOW to H_SEC_LOW+4 =>
								if char_gen (s_index)(v_count-V_SEC_MIN) = '1' then 
									red <= COL_DIGIT; green <= COL_DIGIT; blue <= COL_DIGIT; 
								end if;
								s_index <= s_index+1;
							
							when others =>		-- nothing to do
						end case;		
						sec_col <= sec_col+1;
					end if;

				elsif (v_count <= V_BITS_MAX and v_count >= V_BITS_MIN) or  (v_count <= V_REGS_MAX and v_count >= V_REGS_MIN) then
					-- we are inside the area for the drum data and R/W head display

					-- update drum addresses (track)
					if h_count = h_rez-1 then
						if subline = subline_max then  					-- switch to new track
							adr (16 downto 11) <= std_logic_vector (unsigned(adr(16 downto 11))+1);
						end if;													-- else repeat same track, nothing to do
						subline <= (subline+1) mod (subline_max+1);
					end if;
					
					-- track numbers
					if (subline <= V_TRK_MAX) and (subline >= V_TRK_MIN) then
						case h_count is
						
							when H_TRK_HIGH-1 | H_TRK2_HIGH-1 => 
								-- first digit of track number: get index to character bitmap
								if v_count <= V_BITS_MAX then
									t_index <= char_zero+6*(to_integer(unsigned(adr(16 downto 11)))/10)+1;
								else
									t_index <= 6*CHARACTER'POS(track_label(to_integer(unsigned(adr(13 downto 11))))(0))+1;
								end if;
								
							when H_TRK_LOW-1 | H_TRK2_LOW-1 => 
								-- first digit of track number: get index to character bitmap
								if v_count <= V_BITS_MAX then
									t_index <= char_zero+6*(to_integer(unsigned(adr(16 downto 11))) MOD 10)+1;
								else
									t_index <= 6*CHARACTER'POS(track_label(to_integer(unsigned(adr(13 downto 11))))(1))+1;								
								end if;
								
							when H_TRK_HIGH to (H_TRK_HIGH+4) | H_TRK_LOW to (H_TRK_LOW+4) |
									H_TRK2_HIGH to (H_TRK2_HIGH+4) | H_TRK2_LOW to (H_TRK2_LOW+4) =>
								if char_gen (t_index) (subline-V_TRK_MIN) = '1' then 
									red <= COL_DIGIT; green <= COL_DIGIT; blue <= COL_DIGIT; 
								end if;
								t_index <= t_index+1;	
								
							when others =>
								-- keep the blue value from the background stripe pattern
						end case;				
					end if;

					if (h_count>=H_BITS_LOW) and (h_count<=H_BITS_HIGH) then
					
						-- vertical stripes to delineate word structure
						if (subline/=0) and (subline/=subline_max) then 
							case to_integer (unsigned (adr(4 downto 0))) is
								when 0 			=> blue <= BLUE_OFF;
								when 2 to 7 	=>	blue <= BLUE_SECTOR;
								when 8 to 13	=> blue <= BLUE_TRACK;
								when 16 to 19	=> blue <= BLUE_OPCODE;
								when others		=> blue <= BLUE_OTHER;
							end case;
						end if;
						
						-- data bits from main memory or register track,
						-- dark tick marks (1 pixel wide) for every bit
						
						if (v_count <= V_BITS_MAX) then bit := data;
						else bit := regs (to_integer (unsigned(adr(13 downto 11))));
						end if;
						
						if (subline>0) and (subline<subline_max) then
							if bit = '1' then
								if subcol = 0 then red <= COL_GREY; green <= COL_GREY; blue <= COL_GREY; 
								else red <= COL_WHITE; green <= COL_WHITE; blue <= COL_WHITE;
								end if;
							else
								if subcol = 0 then blue <= COL_OFF;
								end if;
							end if;
						end if;

						-- read/write head
						-- draw green frames around the actual bits read/written

						if animate = '1' then												-- show R/W heads
						
							-- read heads for all tracks
							if ((subline=1 or subline=subline_max-1) and (h_count<H_HEAD_HIGH and h_count>H_HEAD_LOW)) 			-- upper/lower side of frame
								or ((subline>0 and subline<subline_max) and (h_count = H_HEAD_LOW or h_count = H_HEAD_HIGH)) 	-- left or right edge
								or (v_count<=V_BITS_MAX and (adr (16 downto 11)=track) and (subline>0 and subline<subline_max) and (h_count = H_HEAD_LOW-1 OR h_count = H_HEAD_HIGH+1)) 	-- double width left or right edge for active head
							then
								blue <= COL_OFF; red <= COL_OFF;	green <= COL_OFF;	-- R/W head "covers" the data bits
								if (v_count <= V_BITS_MAX) and (adr (16 downto 11) = track) then						-- active track
									if write_en = '1' then red <= COL_WRITE;			-- head is writing
									else green <= COL_READ;									-- head is reading
									end if;
								else green <= COL_PASSIVE;									-- head is here, but passive
								end if;
						
							-- write heads for recirculating registers
							elsif (v_count>=V_REGS_MIN and v_count<=V_REGS_MID) and
								( ((subline=1 or subline=subline_max-1) and (h_count<H_WRHEAD_HIGH and h_count>H_WRHEAD_LOW)) 
								or ((subline>0 and subline<subline_max) and (h_count = H_WRHEAD_LOW or h_count = H_WRHEAD_HIGH)) )
							then
								blue <= COL_OFF; red <= COL_WRITE;	green <= COL_OFF;	-- draw a write head
							end if;
						end if;
					end if;	-- within horizontal area for drum bit display
				
				elsif (showff = '1') then
				
					-- flip-flop status display
					
					-- square "lights"
					if (v_count <= V_FF_MAX) and (v_count >= V_FF_MIN) and (ff_index<15) and (h_count >= ff_pos(ff_index) - FF_SIZE/2) then
						if ((v_count = V_FF_MIN) or (v_count = V_FF_MAX)) and 
							((h_count = ff_pos(ff_index) - FF_SIZE/2) or (h_count = ff_pos(ff_index) + FF_SIZE/2)) then
							-- rounded corners: leave these pixels black
						else
							if ff_state(ff_index) = '1' then				
								-- flip-flop is on
								red <= RED_ON;	green <= GREEN_ON;
							else													
								-- flip-flop is off
								red <= RED_OFF; green <= GREEN_OFF;													
							end if;
						end if;
						
						if (h_count = ff_pos(ff_index) + FF_SIZE/2) then
							-- done displaying this flip-flop
							ff_index <= ff_index+1;
						end if;		
					end if;
					
					-- lettering 
					if (v_count <= V_FFLBL_MAX) and (v_count >= V_FFLBL_MIN) then

						-- lettering inside lights
						if h_count = ff_pos(ff_index)-4 then
							-- begin new character
							f_index <= 6*CHARACTER'POS(ff_label(ff_index));
						end if;

						if (h_count > ff_pos(ff_index)-4) and (h_count < ff_pos(ff_index)+3) then
							-- character output ongoing
							if (char_gen (f_index)(v_count-V_FFLBL_MIN) = '1') then 
								red <= COL_DIGIT; green <= COL_DIGIT; blue <= COL_DIGIT; 
							end if;
							f_index <= f_index+1;
						end if;
				
						-- legends next to lights
						if h_count = FF_POS_PHASE then 	-- phase legend begins here
							if (ff_state(0 to 2) = "000") and (ff_state (6) = '0') then 
								ff_legend <= 9;				-- blocked state
							else 
								ff_legend <= 1 + to_integer (unsigned(ff_state(0 to 2)));
							end if;
							l_char <= 1;
							l_col <= 0;

						elsif h_count = FF_POS_OPCODE then	-- opcode legend begins here
							if (ff_state(0 to 2) /= "000") and  (ff_state(0 to 2) /= "010") then	 
								-- opcode only valid after phase 2
								ff_legend <= 10 + to_integer(unsigned(ff_state(5 to 8))); 
								l_char <= 1;
								l_col <= 0;
							end if;

						elsif h_count = FF_POS_PREG then		-- P register legend begins here
							if  (ff_state(0 to 2) /= "000") and (ff_state(0 to 2) /= "010") and (ff_state(5 to 8) = "1000") then	
								-- valid "print" opcode: argument
								ff_legend <= 27; 
								l_char <= 1;
								l_col <= 0;
							elsif (ff_state(0 to 2) /= "000") and (ff_state(0 to 2) /= "010") and (ff_state(5 to 8) = "0100") then
								-- valid "input" opcode: Input
								ff_legend <= 28; 
								l_char <= 1;
								l_col <= 0;								
							elsif ff_state (2) = '0' then			
								-- phase 1..4: track
								ff_legend <= 26; 
								l_char <= 1;
								l_col <= 0;
							end if;
						end if;
	
						if ff_legend /= 0 then
							if (l_col = 0) then
								-- new character
								l_char <= l_char+1;		-- increment to following character
								l_col <= 6;					-- number of columns in character
								case ff_strings(ff_legend)(l_char) is
									when character'val(0) => 
										-- end of string
										ff_legend <= 0;
									when '#' =>
										-- track number, high digit
										l_index <= char_zero+6*(to_integer(unsigned(ff_state(9 to 14)))/10)+1;
									when '*' =>
										-- track number, low digit
										l_index <= char_zero+6*(to_integer(unsigned(ff_state(9 to 14))) mod 10)+1;
									when others =>
										l_index <= 6*character'pos(ff_strings(ff_legend)(l_char)); 
								end case;
							else		
								-- character output ongoing
								if (char_gen (l_index)(v_count-V_FFLBL_MIN) = '1') then 
									red <= COL_DIGIT; green <= COL_DIGIT; blue <= COL_DIGIT; 
								end if;
								l_col <= l_col-1;
								l_index <= l_index+1;
							end if;
						end if;
					end if;
					
					
					
				end if;

			else			
				-- we are outside the visible area (front porch, sync, back porch) 
            red <= COL_OFF;
            green <= COL_OFF;
            blue <= COL_OFF;
            blank <= '1';
         end if;
			
         -- horizontal sync
			if h_count >= h_sync_start and h_count < h_sync_end then hsync <= '1';      
			else hsync <= '0';
         end if;
         
         -- vertical sync
			if v_count >= v_sync_start and v_count < v_sync_end then vsync <= '1';
         else vsync <= '0';
         end if;
         
			-- update screen position 
         if h_count < h_max then							-- continue in line
            h_count <= h_count+1;
         else
				h_count <= 0;
				ff_index <= 0;
				if v_count < v_max then               	-- new screen line
               v_count <= v_count+1;
				else												
					------ new frame ------
					v_count <= 0;
					
					-- restart drum display 
					subline <= 0;
					adr (16 downto 11) <= (others => '0');
					
					-- set vertical mode constants
					if showff = '1' then 
						V_SEC_MIN  <= 6;												-- sector numbers
						V_SEC_MAX  <= V_SEC_MIN+7;									-- letters are 8 pixels high
						V_BITS_MIN <= 2*(subline_max+1);							-- track data begins. Must be multiple of subline period.
						V_BITS_MAX <= V_BITS_MIN + 64*(subline_max+1) -1;	-- 64 tracks, each 10 pixels high
						V_REGS_MIN <= V_BITS_MAX + 5;								-- track data begins. 
						V_REGS_MID <= V_REGS_MIN + 3*(subline_max+1) -1;	-- 3 tracks with circulating registers
						V_REGS_MAX <= V_REGS_MIN + 6*(subline_max+1) -1;	-- 6 tracks, each 10 pixels high
					else 
						V_SEC_MIN  <= 14;												-- sector numbers
						V_SEC_MAX  <= V_SEC_MIN+7;									-- letters are 8 pixels high
						V_BITS_MIN <= 3*(subline_max+1);							-- track data begins. Must be multiple of subline period.
						V_BITS_MAX <= V_BITS_MIN + 64*(subline_max+1) -1;	-- 64 tracks, each 10 pixels high
						V_REGS_MIN <= V_BITS_MAX + 1*(subline_max+1);		-- track data begins. Must be multiple of subline period.
						V_REGS_MID <= V_REGS_MIN + 3*(subline_max+1) -1;	-- 3 tracks with circulating registers
						V_REGS_MAX <= V_REGS_MIN + 6*(subline_max+1) -1;	-- 6 tracks, each 10 pixels high
					end if;

					-- control display shift
					if animate='1' then
						-- follow LGP-30 drum movement
						-- position R/W head at address 00
--						shift <= 2048 + 32 - HEADPOS - to_integer (unsigned (drumpos));
						shift <= 2048 - HEADPOS - to_integer (unsigned (drumpos));		-- correct now? ###
					else
						-- if screen does not show the position selected via the encoder yet,
						-- "rotate" to the target position gradually
						delta := shift_target - shift;						-- signed distance
						if abs (delta) >= 16 then delta := delta/16;
						elsif delta > 0 then delta := 1;
						elsif delta < 0 then delta := -1;
						end if;
						shift <= shift+delta;
					end if;
					
					-- get stable copy of flip-flop status
					ff_state <= (F, G, H, K, L, Q(1), Q(2), Q(3), Q(4), P(1), P(2), P(3), P(4), P(5), P(6));
				end if;
         end if;

      end if;
   end process;
	
	test(0) <= hsync;
	test(1) <= '1' when drumpos = "00000000000" else '0' ;
	test(2) <= blank;
	test(3) <= blue(7);
	test(4) <= '0';
	test(5) <= '0';
	test(6) <= '0';
	test(7) <= '0';
	
end Behavioral;

