----------------------------------------------------------------------------------

-- LGP-30 drum computer simulation
-- LCD as alternative for CRT display (register output - three traces on 3*16 char display)
--
-- jm 25.8.2016

-- Electronic Assembly EA DOGM132, 32*132 pixels,
-- ST7565R display controller with SPI bus.
-- Must be initialized (by lcd_controller) to have page 0, col 0 in the upper left corner.
--
-- Order of displayed traces (bottom to top) is A, R, C.
--
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

library work;
use work.constants.all;

entity lcd is
	port (
		-- all registers are presented during each word cycle
		A: in std_logic;			-- accumulator to display
		R: in std_logic;			-- instruction register to display
		C: in std_logic;			-- program counter to display
		
		-- bit clock and trigger
		-- triggers horizontal deflection (right to left!) and 3-step counter for vertical offset
		T: in std_logic;			-- bit clock
		t3: in std_logic;			-- high during sign bit cycle
		io_clk: in std_logic;	-- 10 MHz
		
		-- display settings 
		ext_disp: in std_logic;			-- extended display on?
		cont_up: in std_logic;			-- increase contrast value (1 step)
		cont_dn: in std_logic;			-- decrease contrast value (1 step)

		-- flip-flops for status display
		F, G, H, K, L: in std_logic;
		P: in std_logic_vector (1 to 6);
		Q: in std_logic_vector (1 to 4);
		
		-- clock rate display
		inv_rate: 	in std_logic;					-- invert clock rate display?
		cpow: 		in freq_index; 				-- slowing clock down by 2^cpow; 0 = 125 kHz

		-- output to external display.
		-- Electronic Assembly EA DOGM132, 32*132 pixel,
		-- ST7565R display controller, SPI interface.
		si				: OUT STD_LOGIC;  			-- data
		scl			: OUT	STD_LOGIC;				-- clock
		cs_n			: OUT STD_LOGIC;				-- chip select, active low
		a0				: OUT STD_LOGIC;				-- high = data, low = command
		res_n			: OUT STD_LOGIC				-- reset, active low
	);
end lcd;

architecture Behavioral of lcd is

component lcd_controller 
	PORT(
		-- connections to client in FPGA
		clk        : IN    STD_LOGIC;  -- system clock
		reset_n    : IN    STD_LOGIC;  -- active low reinitializes lcd
		enable     : IN    STD_LOGIC;  -- latches data into lcd controller
		data		  : IN    STD_LOGIC_VECTOR(8 DOWNTO 0);  -- data and control signals
		busy       : OUT   STD_LOGIC := '1';  -- lcd controller busy/idle feedback

		-- connections to physical LCD
		si				: OUT STD_LOGIC;  			-- data
		scl			: OUT	STD_LOGIC;				-- clock
		cs_n			: OUT STD_LOGIC;				-- chip select, active low
		a0				: OUT STD_LOGIC;				-- high = data, low = command
		res_n			: OUT STD_LOGIC				-- reset, active low
	); 
end component;

-- communication with LGP-30 logic unit
signal Aser, Rser, Cser: std_logic_vector (31 downto 0);					-- shift registers
type regtable is array (3 to 10) of std_logic_vector (31 downto 0);	-- parallel buffers
signal reg: regtable;

-- communication with LCD low-level control
signal lcd_enable: std_logic := '0';
signal lcd_busy: std_logic;
signal lcd_data: std_logic_vector(8 downto 0);

-- state machine for trace output
constant NPAGES: natural := 11;							-- n umber of display pages (lines)
signal pix_counter: natural range 0 to 3 := 0;		-- repeat every bit for x pixels
signal bit_counter: integer range -1 to 31 := 0;	-- current bit in register; counts up 		
signal page: natural range 0 to NPAGES-1 := 0;		-- active display line; counts up

-- state control for clear screen operation
signal clear_page: natural range 0 to 3 := 0;		-- count physical display pages to be cleared
signal clear_col: natural range 0 to 136 := 1;		-- start with a clear screen operation
signal was_ext: std_logic := '1';						-- output mode so far had extended information

-- LCD output patterns
-- Four sets, depending on vertical orientation within the display page:
--
-- 0 = bottom
-- 1 = centered
-- 2 = upper half
-- 3 = lower half
--
type bit_pattern is array (0 to 3) of std_logic_vector(8 downto 0);
constant LCD_LOW:  bit_pattern := ("111000000", "101100000", "100000000", "100000110");
constant LCD_HIGH: bit_pattern := ("100110000", "100000110", "101100000", "100000000");
constant LCD_CHNG: bit_pattern := ("100000000", "101111110", "111100000", "100000111");	-- vertical lines all the way up to min/max bit levels
-- constant LCD_CHNG: bit_pattern := ("100000000", "100111100", "111000000", "100000011");	-- vertical lines 1 pixel shorter ("low pass" on trace)
constant LCD_GAP:  bit_pattern := ("100000000", "100000000", "100000000", "100000000");
constant LCD_DOT:  bit_pattern := ("100000000", "100000000", "100000100", "101000000");	-- small tick marks (delineate bits)
constant LCD_TICK: bit_pattern := ("100000000", "100000000", "100000110", "101100000");	-- large tick marks (delineate bit groups)

--  vertical orientation of bit patterns depends on the display page
type vert_lookup is array (natural range 0 to NPAGES-1) of natural range 0 to 3;
constant vert_mode: vert_lookup := 
(
	0, 
	-- extended display; all bit patterns at bottom of respective line
	0, 0, 0, 0, 0, 	
	-- standard display: C centered, R split upper/lower, A centered
	0, 1, 2, 3, 1
);

-- contrast setting
constant CONTRAST_INI: natural := 31;
signal contrast: natural range 0 to 255 := CONTRAST_INI;

-- commands
constant LCD_PAGE: natural := 2#10110000#;
constant LCD_COLH: natural := 2#00010000#;
constant LCD_COLL: natural := 2#00000000#;
constant LCD_CONT: natural := 2#10000001#;
constant LCD_NOP:  natural := 2#11100011#;

-- command table for start of pages
signal cmd_index: natural range 0 to 3*NPAGES-1 := 0;		-- index to cmd_list table
signal cmd_col: natural range 0 to 3;							-- count down bytes until command complete

type cmd_map is array (0 to 3*NPAGES-1) of natural range 0 to 255;
signal cmd_list: cmd_map :=
(
	LCD_CONT, CONTRAST_INI, LCD_NOP,			-- 0: contrast setting. note that CONTRAST_INI will be overwritten later

	-- extended display mode with flip-flop display
	LCD_PAGE+0, LCD_COLH+0, LCD_COLL+0,		-- 1: flip-flops
	LCD_PAGE+1, LCD_COLH+0, LCD_COLL+0,		-- 2: CPU clock frequency
	LCD_PAGE+1, LCD_COLH+4, LCD_COLL+10,	-- 3: C register
	LCD_PAGE+2, LCD_COLH+0, LCD_COLL+2,		-- 4: R register
	LCD_PAGE+3, LCD_COLH+0, LCD_COLL+2,		-- 5: A register
	
	-- standard display mode
	LCD_PAGE+0, LCD_COLH+0, LCD_COLL+0,		-- 6: CPU clock frequency
	LCD_PAGE+0, LCD_COLH+4, LCD_COLL+10,	-- 7: C register
	LCD_PAGE+1, LCD_COLH+0, LCD_COLL+2,		-- 8: R register, upper half
	LCD_PAGE+2, LCD_COLH+0, LCD_COLL+2,		-- 9: R register, lower half
	LCD_PAGE+3, LCD_COLH+0, LCD_COLL+2		-- 10: A register
);

-- character generator

signal char_index: natural range 0 to 6*128-1 := 0;		-- index to char_gen table
signal char_col: natural range 0 to 7;							-- count down colummns until character complete
signal char_inverse: std_logic := '0';							-- print character inverted?

begin

	-- low-level LCD controller
	controller_inst: lcd_controller
		port map (
			clk => io_clk,
			reset_n => '1',
			enable => lcd_enable,
			data => lcd_data,
			busy => lcd_busy,

			si => si,
			scl => scl,
			cs_n => cs_n,
			a0 => a0,
			res_n => res_n
		);


	-- serial to parallel conversion and buffering of A, R, C registers.
	-- reg table has same numbering as the display pages
	process (T)
	variable i: natural range 0 to 31 := 0;
	begin
		if rising_edge (T) then
			if t3 = '1' then i := 0;
			else i := i-1;
			end if;
			reg (3)(i) <= C;
			reg (4)(i) <= R;
			reg (5)(i) <= A;
			reg (7)(i) <= C;
			reg (9)(i) <= R;
			reg (8)(i) <= R;
			reg (10)(i) <= A;
		end if;
	end process;
	
	
	-- drive LCD 
	process (io_clk)
	variable lcd_cache: std_logic_vector(8 downto 0);	-- cache for tick marks while building display line
	begin
		if rising_edge (io_clk) then

			-- monitor encoder (contrast up/down)
			-- limit contrast range (1..63) somewhat, to keep the display readable
			if (cont_up='1') and (contrast<47) then contrast <= contrast+1;
			elsif (cont_dn='1') and (contrast>15) then contrast <= contrast-1;
			end if;
			cmd_list (1) <= contrast;
		
			if lcd_enable = '1' then			
				----
				---- reset enable after previously sent character
				----
				lcd_enable <= '0';				-- have just sent a byte - don't do anything except reset the trigger
			elsif lcd_busy = '0' then			-- lcd is ready to receive data
			
				if cmd_col /= 0 then
					----
					---- ongoing command sequence output (start of new page)
					----
					lcd_data (7 downto 0) <= std_logic_vector (to_unsigned (cmd_list (cmd_index), 8));
					lcd_data (8) <= '0';			-- command
					lcd_enable <= '1';			-- output the word
					cmd_index <= cmd_index+1;
					cmd_col <= cmd_col-1;
		
				elsif char_col /= 0 then
					----
					---- ongoing character output 
					----
					if char_col = 1 then			-- add spacer column and finish letter
						if char_inverse = '1' then lcd_data (7 downto 0) <= "11111111";
						else lcd_data (7 downto 0) <= "00000000";	
						end if;
					else					
						if char_inverse = '1' then lcd_data (7 downto 0) <= NOT char_gen (char_index);
						else lcd_data (7 downto 0) <= char_gen (char_index);					
						end if;
					end if;
					lcd_data (8) <= '1';			-- data not command
					lcd_enable <= '1';			-- output the word
					char_index <= char_index+1;
					char_col <= char_col-1;

				elsif clear_col /= 0 then
					----
					---- clear screen in progress
					----
					clear_col <= clear_col+1;					-- next column (at next clock - unless changed below
					case clear_col is
						when 1 => 
							lcd_data (7 downto 0) <= std_logic_vector (to_unsigned (LCD_PAGE + clear_page, 8));
							lcd_data (8) <= '0';
						when 2 => 
							lcd_data (7 downto 0) <= std_logic_vector (to_unsigned (LCD_COLH, 8));
							lcd_data (8) <= '0';
						when 3 => 
							lcd_data (7 downto 0) <= std_logic_vector (to_unsigned (LCD_COLL, 8));
							lcd_data (8) <= '0';
						when 4 to 135 => 
							lcd_data (7 downto 0) <= (others => '0');
							lcd_data (8) <= '1';
						when others =>								-- this page is cleared
							if clear_page=3 then
								clear_col <= 0;					-- all pages cleared
								page <= 0;							-- continue with contrast setting
								cmd_index <= 0;
							else
								clear_page <= clear_page+1;	-- next page
								clear_col <= 1;	
							end if;
					end case;
					lcd_enable <= '1';							-- output the word

				elsif page = 0 then
					---
					--- contrast setting
					---
					
					-- contrast setting command has been sent, switch directly to next page.
					-- gate here depending on display mode (extended or standard)
					if ext_disp = '1' then			-- extended display: show flip-flops 
						page <= 1;
						cmd_index <= 3*1;
					else
						page <= 6;						-- no extended: show clock rate display
						cmd_index <= 3*6;
					end if;
					bit_counter <= 0;					-- start with first character
					cmd_col <= 3;

				elsif page = 1 then
					----
					---- flip-flop status line, with character display
					----
					char_col <= 7;
					bit_counter <= bit_counter+1;
					case bit_counter is
					
						when  0 =>	char_index <= 6*70; char_inverse <= F;	
						when  1 =>	char_index <= 6*71; char_inverse <= G;
						when  2 =>	char_index <= 6*72; char_inverse <= H;
						when  3 =>	char_index <= 6*75; char_inverse <= K;
						when  4 =>	char_index <= 6*76; char_inverse <= L;
						when  5 =>	char_index <= 6*32; char_inverse <= '0';	-- ' '

						when  6 =>	char_index <= 6*81; char_inverse <= '0';	-- 'Q'
						when  7 =>	char_index <= 6*49; char_inverse <= Q(1);	-- '1'
						when  8 =>	char_index <= 6*50; char_inverse <= Q(2);	-- '2'
						when  9 =>	char_index <= 6*51; char_inverse <= Q(3);	-- '3'
						when 10 =>	char_index <= 6*52; char_inverse <= Q(4);	-- '4'
						when 11 =>	char_index <= 6*32; char_inverse <= '0';	-- ' '

						when 12 =>	char_index <= 6*80; char_inverse <= '0';	-- 'P'
						when 13 =>	char_index <= 6*49; char_inverse <= P(1);	-- '1'
						when 14 =>	char_index <= 6*50; char_inverse <= P(2);	-- '2'
						when 15 =>	char_index <= 6*51; char_inverse <= P(3);	-- '3'
						when 16 =>	char_index <= 6*52; char_inverse <= P(4);	-- '4'
						when 17 =>	char_index <= 6*53; char_inverse <= P(5);	-- '5'
						when 18 =>	char_index <= 6*54; char_inverse <= P(6);	-- '6'
					
						when others => 							-- end of text line
							char_col <= 0;							-- no more text output
							bit_counter <= 0;						-- restart with first character (text line follows)
							page <= 2;								-- next display line: clock rate
							cmd_index <= 3*2;						-- set page and column address 
							cmd_col <= 3;
					end case;

				elsif (page = 2) or (page = 6) then
					----
					---- clock frequency output
					----
					char_inverse <= inv_rate;
					if bit_counter<6 then
						char_index <= 6*CHARACTER'POS(freq_list(cpow)(bit_counter));
						char_col <= 7;
						bit_counter <= bit_counter +1;
					else
						page <= page+1;
						cmd_index <= 3*(page+1);
						cmd_col <= 3;
						bit_counter <= 17;							-- C display next: skip 18 high bits
						pix_counter <= 0;
					end if;
							
				else
					----
					---- register output (C, R, A)
					----
					
					-- check for change of display mode - need to clear screen?
					if (ext_disp /= was_ext) then
						-- no more flip-flop and clock display - clear the screen once
						clear_col <= 1;
						clear_page <= 0;
					end if;					
					was_ext <= ext_disp;
					
					pix_counter <= pix_counter-1;
					if pix_counter = 0 then				
											
						-- start with new bit
						
						-- add scale ticks between the oscilloscope traces
						-- (only in standard display mode, which provides extra height per trace)
						if (page=9) then
							-- groups of 4 bits, between A and R register
							if (bit_counter mod 4 = 3) then lcd_cache := LCD_TICK (vert_mode (page));
--							else lcd_cache := LCD_DOT (vert_mode (page));
							else lcd_cache := LCD_GAP (vert_mode (page));	-- no dots here, too confusing
							end if;
						elsif (page=8) then
							-- delineate opcode, track, sector fields between R and C
							if (bit_counter = 29 or bit_counter = 23 or bit_counter = 17 or bit_counter = 15 or bit_counter = 11) then
								lcd_cache := LCD_TICK (vert_mode (page));
							elsif (bit_counter < 29 and bit_counter > 17) or (bit_counter < 15 and bit_counter > 11) then
								lcd_cache := LCD_DOT (vert_mode (page));
							else 
								lcd_cache := LCD_GAP (vert_mode (page));
							end if;
						else
							lcd_cache := LCD_GAP (vert_mode (page));
						end if;
						lcd_data <= lcd_cache;
						
						-- check for beginning of new line
						if (bit_counter=31) or ((page=3 or page=7) and bit_counter=29) then	-- have completed a register output
							-- new line
							cmd_col <= 3;
							bit_counter <= -1;					-- restart with high bit
							pix_counter <= 0;
							if (page=10) or (page=5) then 	-- display content complete, restart output cycle
								page <= 0;
								cmd_index <= 0;
							else
								page <= page+1;					-- next display line
								cmd_index <= 3*(page+1);		-- cmd output due: set page and column address 
							end if;
						else
							-- same line, next bit: add vertical lines between bit changes
							if reg (page)(bit_counter) /= reg (page)(bit_counter+1)then
								lcd_data <= lcd_cache or LCD_CHNG (vert_mode (page));	-- bit state will change
							end if;
							bit_counter <= bit_counter+1;
						end if;
						
					else
						-- keep displaying the current bit
						if reg (page)(bit_counter) = '1' then
							lcd_data <= LCD_HIGH (vert_mode (page));
						else
							lcd_data <= LCD_LOW (vert_mode (page));
						end if;
					end if;
					
					lcd_enable <= '1';					-- output the word
					
				end if;					
			end if;
		end if;
	end process;

end Behavioral;

