----------------------------------------------------------------------------------
--
-- LGP-30 drum computer simulation
-- Logic module (flip-flops and diode matrix)
--
-- jm 29.7.2016
-- jm 17.10.2017 fixed Qout(4) output
--
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity logic is
	port (
	
			-- system clock
			clk: in std_logic;
			clk_en: in std_logic;					-- clock enable (disabled during xmodem transfers)
			
			-- main memory
			V, Vn: in std_logic;						-- read head
			Vp: out std_logic;						-- write head
			W: buffer std_logic;						-- write enable
			P: out std_logic_vector (1 to 6);	-- track selector
		
			-- recycling registers
			A, An: in std_logic;						-- accumulator read, 32 bits behind write
			As, Asn: in std_logic;					-- accumulator read, 65 bits behind write
			Ap: out std_logic;						-- accumulator write, always enabled

			R, Rn: in std_logic;						-- instruction read, 32 bits behind write
			Rp: out std_logic;						-- instruction write, always enabled

			C, Cn: in std_logic;						-- program counter read, 32 bits behind write
			Cp: out std_logic;						-- program counter write, always enabled
			
			-- timing signals
			S1, S1n: in std_logic;
			S2, S2n: in std_logic;
			S3, S3n: in std_logic;

			-- control panel input
			bs: in std_logic;							-- start
			bq, bqn: in std_logic;					-- manual input
			bcn: in std_logic;						-- clear counter
			br, brn: in std_logic;					-- fill instruction
			Ps, Psn: in std_logic;					-- execute instruction
			O1, O1n: in std_logic;					-- single step

			T: in std_logic_vector (1 to 4);		-- breakpoint switches
			D, Dn: in std_logic;						-- select 6 bit input
			Tc: in std_logic;							-- transfer control
			
			Tbs: out std_logic;						-- blocked state flag - drives stop/compute lights

			-- typewriter
			Tp: out std_logic_vector (1 to 6);			-- character to be printed/punched
			PPs, PPsn: in std_logic_vector (1 to 6);	-- character from keyboard or tape reader
			Tx: out std_logic;								-- trigger printing
			Tread: out std_logic;							-- trigger reading of next 4/6 bit input
			Tdelay: out std_logic;							-- trigger delay in Flexwriter (not needed in simulation?)
			
			-- oscillosope display: trigger
			t3, t3n: buffer std_logic;						-- high during sign bit cycle
			
			-- outputs to LCD
			Fout, Gout, Hout, Kout, Lout: out std_logic; -- phase bits
			Qout: out std_logic_vector (1 to 4);		-- instruction decoding
			
			-- debug outputs
			test: out std_logic_vector (7 downto 0)

	);
	
end logic;

architecture Behavioral of logic is

------------------------------------------------
-- active logic units (LGP-30 plug-in boards) declared explicitly:
-- R/S flip-flop and Inverter
------------------------------------------------

	component rs_ff 
		port(
			clk: std_logic;			-- clocks flip-flop on rising edge
			clk_en: in std_logic;	-- clock enable
			R, S: in std_logic;		-- input reset & set
			Q, Qn: out std_logic		-- output and inverted output
		);
	end component;

	component inv 
		port(
			D: in std_logic;			-- input 
			Dn: out std_logic			-- output(inverted input)
		);
	end component;

------------------------------------------------
-- signals within logic block
------------------------------------------------

-- flip-flop connections: output and inverted output, set and reset input 

	signal F, Fn, Fp, Fpn: std_logic;
	signal G, Gn, Gp, Gpn: std_logic;
	signal H, Hn, Hp, Hpn: std_logic;
	signal K, Kn, Kp, Kpn: std_logic;
	signal L, Ln, Lp, Lpn: std_logic;
	
	signal P1, P1n, P1p, P1pn: std_logic;
	signal P2, P2n, P2p, P2pn: std_logic;
	signal P3, P3n, P3p, P3pn: std_logic;
	signal P4, P4n, P4p, P4pn: std_logic;
	signal P5, P5n, P5p, P5pn: std_logic;
	signal P6, P6n, P6p, P6pn: std_logic;

	signal Q1, Q1n, Q1p, Q1pn: std_logic;
	signal Q2, Q2n, Q2p, Q2pn: std_logic;
	signal Q3, Q3n, Q3p, Q3pn: std_logic;
	signal Q4, Q4n, Q4p, Q4pn: std_logic;

-- inverter connections
	
	signal r1, r1n: std_logic;
	signal S, Sn: std_logic;
	signal Vpn: std_logic;			-- Vp declared externally
	signal I1, I1n: std_logic;
	signal I2, I2n: std_logic;
	
-- internal terms ("alpha" terms in LGP-30 documentation)
-- Signal declaration should work here, without creating a buffer,
-- since these signals are in the sensitivity list of the process

	signal a1, a2, a3, a4, a5, a6: std_logic;
	signal a7, a8, a9, a10, a11, a12: std_logic;
	signal a13, a14, a15, a16, a17, a18: std_logic;
	
	signal i: std_logic;
	signal X, Y: std_logic;
	
begin

------------------------------------------------
-- flip-flop instances
------------------------------------------------

	F_ff: rs_ff		-- operation phase, lowest bit
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	Fpn,
			S =>	Fp,
			Q =>	F,
			Qn =>	Fn
		);
		
	G_ff: rs_ff		-- operation phase, middle bit
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	Gpn,
			S =>	Gp,
			Q =>	G,
			Qn =>	Gn
		);
		
	H_ff: rs_ff		-- operation phase, highest bit
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	Hpn,
			S =>	Hp,
			Q =>	H,
			Qn =>	Hn
		);
		
	K_ff: rs_ff		-- sector match
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	Kpn,
			S =>	Kp,
			Q =>	K,
			Qn =>	Kn
		);
		
	L_ff: rs_ff		-- math carry
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	Lpn,
			S =>	Lp,
			Q =>	L,
			Qn =>	Ln
		);

	P1_ff: rs_ff		-- track address bit 5, typewriter input 
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	P1pn,
			S =>	P1p,
			Q =>	P1,
			Qn =>	P1n
		);
		
	P2_ff: rs_ff		-- track address bit 4, typewriter input 
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	P2pn,
			S =>	P2p,
			Q =>	P2,
			Qn =>	P2n
		);
		
	P3_ff: rs_ff		-- track address bit 3, typewriter input 
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	P3pn,
			S =>	P3p,
			Q =>	P3,
			Qn =>	P3n
		);
		
	P4_ff: rs_ff		-- track address bit 2, typewriter input 
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	P4pn,
			S =>	P4p,
			Q =>	P4,
			Qn =>	P4n
		);
		
	P5_ff: rs_ff		-- track address bit 1, operand sign bit
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	P5pn,
			S =>	P5p,
			Q =>	P5,
			Qn =>	P5n
		);
		
	P6_ff: rs_ff		-- track address bit 0, operand sign, mult/div intermediate storage
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	P6pn,
			S =>	P6p,
			Q =>	P6,
			Qn =>	P6n
		);
		
	Q1_ff: rs_ff		-- opcode, bit 3
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	Q1pn,
			S =>	Q1p,
			Q =>	Q1,
			Qn =>	Q1n
		);

	Q2_ff: rs_ff		-- opcode, bit 2
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	Q2pn,
			S =>	Q2p,
			Q =>	Q2,
			Qn =>	Q2n
		);

	Q3_ff: rs_ff		-- opcode, bit 1
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	Q3pn,
			S =>	Q3p,
			Q =>	Q3,
			Qn =>	Q3n
		);

	Q4_ff: rs_ff		-- opcode, bit 0
		PORT MAP (
			clk => clk,
			clk_en => clk_en,
			R =>	Q4pn,
			S =>	Q4p,
			Q =>	Q4,
			Qn =>	Q4n
		);

------------------------------------------------
-- inverter instances
------------------------------------------------

	r1_inv: inv		-- 
		PORT MAP (
			D =>	r1,
			Dn =>	r1n
		);

	S_inv: inv		-- 
		PORT MAP (
			D =>	S,
			Dn =>	Sn
		);

	t3_inv: inv		-- sign bit gate
		PORT MAP (
			D =>	t3,
			Dn =>	t3n
		);

	Vpn_inv: inv		-- next memory bit to write
		PORT MAP (
			D =>	Vpn,
			Dn =>	Vp
		);

	I1_inv: inv		-- 
		PORT MAP (
			D =>	I1,
			Dn =>	I1n
		);

	I2_inv: inv		-- 
		PORT MAP (
			D =>	I2,
			Dn =>	I2n
		);

------------------------------------------------
-- logic equations
------------------------------------------------

	Fp 	<= (Ps and t3) or (Fn and (a14 or (Gn and a3))); 
	Fpn 	<= a17 or (t3 and P1n and a13) or (F and a15 and K and Q4n); 
	
	Gp		<= Gn and ((K and Psn and ((F and a16) or (Q2 and a16) or (Q4 and a15))) or (a18 and F) or (a2 and a10 and F and Psn and t3));
	Gpn	<= a14 or (G and t3 and (Fn or (P1n and Q3)));
	
	Hp		<= (a17 and a2 and a9);
	Hpn	<= (a12 and t3) or (a15 and ((K and Q4n) or (G and P1n and Q3)));
	
	Kp		<= t3;
	Kpn	<= (S2 and Hn and ((G and Cn) or (Gn and S3n and ((S1n and r1) or (S1 and r1n)))))
				or (S2n and H and ((S1n and r1) or (S1 and S3n and r1n)));
				
	Lp		<= W or ((a5 or a7) and (t3n or (H and Q3 and P1n))) or (a13 and Q2n);
	Lpn	<= a16 or (P1 and t3) or (Q3n and t3) or ((a6 or a8) and (Q1n or Q3) and Q2);
	
	P1p	<= (i and Gn and r1) or (i and G and A) or (a15 and P1n) or PPs(1);						
	P1pn	<= (i and Gn and r1n) or (i and G and An) or a3 or (G and t3 and P1) or PPsn(1);		
	
	P2p	<= (i and P1)  or PPs(2);
	P2pn	<= (i and P1n) or PPsn(2);
	
	P3p	<= (i and P2)  or PPs(3);
	P3pn	<= (i and P2n) or PPsn(3);
	
	P4p	<= (i and P3)  or PPs(4);
	P4pn	<= (i and P3n) or PPsn(4);
	
	P5p	<= (i and P4)  or (a14 and V)  or PPs(5);
	P5pn	<= (i and P4n) or (a14 and Vn) or PPsn(5);
	
	P6p	<= (i and P5)  or (a17 and A)  or (a3 and ((A and Q3n)  or (As and Q3)))  or PPs(6);
	P6pn	<= (i and P5n) or (a17 and An) or (a3 and ((An and Q3n) or (Asn and Q3))) or PPsn(6);
	
	Q1p	<= (a11 and R);
	Q1pn	<= (a11 and Rn) or bq;
	
--	Q2p	<= (Fn and Gn and Hn and bs) or (a11 and Q1) or 
--				(a13 and t3 and Q2n and O1n and (Q1 or a9 or (P1 and T(1)) or (P2 and T(2)) or (P3 and T(3)) or (P4 and T(4)))) or (bq and Ps);
	Q2p	<= (Fn and Gn and Hn and bs) or (a11 and Q1) or 
				(Hn and a13 and t3 and Q2n and O1n and (Q1 or a9 or (P1 and T(1)) or (P2 and T(2)) or (P3 and T(3)) or (P4 and T(4)))) or (bq and Ps);
	Q2pn	<= (a11 and Q1n) or (Fn and Gn and a3 and Q3n and ((A and P6) or (An and P6n)))
				or (G and t3 and ((a2 and a10) or (O1 and Q1) or (Q1 and Q2 and Q3 and ((Ln and a7) or (Ln and a5) or (L and a6) or (L and a8))) )) or (O1 and Hpn);
	-- Note that Q2pn depends on Hpn (i.e. the flip-flop input!) watch out for timing/latch issues! 
	-- Should work, since Hpn is in the process' sensitivity list?		
	-- Seems OK, from looking at the RTL schematic.
	
	Q3p	<= a11 and Q2;
	Q3pn	<= (a11 and Q2n) or bq;
	
	Q4p	<= a11 and Q3;
	Q4pn	<= (a11 and Q3n) or bq or (a18 and (A or (R and Tc)));
			
	-- Thyratrons
	
	Tp(1)	<= X and P1;
	Tp(2)	<= X and P2;
	Tp(3)	<= X and P3;
	Tp(4)	<= X and P4;
	Tp(5)	<= X and P5;
	Tp(6)	<= X and P6;
	Tx		<= X and Y;
	Tread	<= X and P1n and P2n and P3n and P4n and P5n and P6n;	-- prepare printer for input (P0 order)
	Tdelay <= X and P1n and P2 and P5n and P6n;			-- printer needs delay (e.g. after carriage return)
	Tbs	<= F or G or Q2;		-- "blocked state" flag drives control panel status lights

	-- Inverter inputs
	
	r1		<= (C and Fn) or (C and H) or (F and Hn and R);
	S		<= (Hn and Q4) or a12 or (H and Q3n and a4) or (Fn and H and Q3);
	t3 	<= S1 and S3;
	Vpn	<= (Q1n and Q4 and ((Kn and Cn) or (K and C))) or (An and (Q1 or Q4n)) or Ln;

	I1		<= (a12 and Asn and P5n) or ((A and (Hn or (a13 and Q3 and P1n))) or (Fn and Gn and P1n and Q3n and (A or (t3 and P6)))
				or (As and H and ((F and Gn) or (Q3 and P1) or (Q3n and (F or (Gn and P1) or (G and P5))))));
	I2		<= (a12 and a4) or (Hn and V) or (H and R and P1n and (Q3n or (Gn and P6))) 
				or (H and Q3 and ((G and A and P1n) or (P1 and P5 and (Fn or P6))));

	-- Cathode follower inputs

	i		<= (a13 and Hn and a2 and a10) or (S2 and S3 and Hn and Gn and (Q1 or Q2n or a9 or Fn));
	X		<= a13 and a10 and Q1 and Q2n;		-- trigger output thyratrons. 
	W		<= (a1 and Q3n) or (Q1n and Q2n and Q3 and S2 and a13);

-- 	TEST: simplified circulation
--		Rp <= R or t3;
--		Ap	<= A or t3;
	
	Rp		<= (br and A) or (G and Hn and V and (Fn or (a2 and a9))) or (brn and R and (H or Gn or (F and (Q1 or Q2n or a10))));
	Cp		<= (Hn and ((a13 and R and Q1 and Q2n and Q3 and Q4n) or (Gn and S1 and S2n and S3n) or (Fn and G and S2 and ((K and Cn) or (Kn and C)))))
				or (bcn and C and (H or (G and S2n) or (Gn and S2) or (F and S2 and (Q1n or Q2 or Q3n or Q4))));

	Ap		<= (Hn and 
					(
						(a13 and ( (a2 and a10 and ((P4 and Dn) or (P6 and D))) or (Q2n and Q3n and Q4 and V and (Q1n or A)) ))
						or 
						(A and ((Q1 and a10) or Fn or Gn or (Q2n and (Q3 or Q4n)) or (a2 and a9 and t3n)))
					)
				)
				or 
				(
					((a1 and Q3) or (H and t3n) or (H and P1n) or a12) 
					and 
					((L and I1 and I2) or (L and I1n and I2n) or (Ln and I1 and I2n) or (Ln and I1n and I2))
				);
	Y		<= (P1 or P2 or P3 or P4 or P5 or P6);
			
	-- common partial terms
	
	a1		<= F and G and Q1 and Q2;
	a2		<= Q1n and Q2;
	a3		<= H and P1 and t3;
	a4		<= (P5 or P6n) and (P5n or P6);
	a5		<= I1 and I2 and Sn;
	a6		<= I1 and I2n and S;
	a7		<= I1n and I2 and S;
	a8		<= I1n and I2n and Sn;
	a9		<= Q3 or Q4;
	a10	<= Q3n and Q4n;
	a11	<= Hn and Gn and F and S1n and S2n and S3 and bqn;
	a12	<= Fn and G and H;
	a13	<= F and G;
	a14	<= G and Hn and S1 and S3;
	a15	<= H and S1 and S3;
	a16	<= Hn and S1 and S3;
	a17	<= F and G and Hn and S1 and S3;
	a18	<= Q1 and Q2n and Q3 and S1 and S3;

	P(1)	<= P1;
	P(2)	<= P2;
	P(3)	<= P3;
	P(4)	<= P4;
	P(5)	<= P5;
	P(6)	<= P6;
		
	------------------------------------------------
	-- outputs to LCD only
	------------------------------------------------	
	
	Qout(1)	<= Q1;
	Qout(2)	<= Q2;
	Qout(3)	<= Q3;
	Qout(4)	<= Q4;
	
	Fout	<= F;
	Gout 	<= G;
	Hout	<= H;
	Kout	<= K;
	Lout	<= L;
	
	------------------------------------------------
	-- assign test outputs for debugging
	------------------------------------------------
	
	test(0)	<= Q1;
	test(1)	<= Q2;
	test(2)	<= Q3;
	test(3)	<= Q4;
	test(4)	<= i;
	test(5)	<= Ps;
	test(6)	<= P1;
	test(7)	<= P2;
	
end Behavioral;

