----------------------------------------------------------------------------------

-- LGP-30 drum computer simulation
-- Operating panel
--
-- jm 1.8.2016


-- Controls and indicators, according to the LGP-30 documentation:
--
-- Power control - don't simulate here:
--		POWER ON
--		POWER OFF
--		STAND BY
--		OPERATE
--		STAND BY TO OPERATE (indic ator light only)
--
-- Status indicator lights only - simulate via LEDs:
--		STOP
-- 	COMPUTE
--
-- Mode buttons:
-- Probably mechanically interlocked against each other (press one to release the others).
-- Light indicates which mode is active.
-- Simulate via pushbutton and LED, with interlock logic in FPGA.
--		MANUAL INPUT
--		ONE OPERATION
--		NORMAL
--
-- Operation buttons:
-- Not latching, push once to trigger an action.
-- Indicator light on each button shows whether it is enabled.
-- Simulate via pushbutton and LED.
-- Pulse length emitted by these buttons depends on the LGP-30 clock!
--		START
--		CLEAR COUNTER
--		FILL INSTRUCTION
--		EXECUTE INSTRUCTION
--
-- Latching pushbuttons:
--	Push to activate, push again to release.
-- Indicator light shows whether it is active.
-- Simulate via pushbutton and LED, with latching logic in FPGA.
--		4* BREAK POINT
--		6 BIT INPUT
--		TRANSFER CONTROL
--		TTY Start/stop read
--		TTY Connect
--
--	For simulator control:
-- 	2* Encoder X/Y 
--		Encoder pushbutton + LED 
--
-- Total I/O needs:
--		16 pushbuttons + LEDs
--		2 additional status LEDs
--		2 additional encoder inputs
--
-- External switch/LED connections:
-- 	0	T0		break point 32
-- 	1	T1		break point 16
-- 	2	T2		break point  8
-- 	3	T3		break point  4
-- 	4	D		6-bit input 
--		5	Tc		transfer control
--
--		6	bs		Start	
--		7	bc		Clear
--		8	br		Fill
--		9 	Ps		Exec
--
--	  10	Nor	normal mode
--	  11	Sin	single step mode
--	  12	Man	manual mode
--
--	  13	TTY start/stop read
--   14	TTY connect
--
--	  15  Encoder button + LED
--	  16  Spare button + LED
--
--	  18  Mimas SW4+LED - option
--	  19  Mimas SW3+LED - option
--	  20  Mimas SW2+LED - option
--	  21  Mimas SW1+LED - option
-- 
--	  21	Comp status LED + option button
--   22	Stop status LED + option button

--	  23	Encoder X
--   24	Encoder Y

-- Note that actuated switches read as '0'.
-- Encoder switches must be in last position, since they are excluded from the general debouncing scheme
-- (need faster time constant and different logic, implemented in the encoder component)
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

library work;
use work.constants.all;


entity panel is
	port (	

		io_clk: in std_logic;						-- 5 Mhz I/O clock; for debounce timer only
		lgp_clk: in std_logic;						-- LGP-30 bit clock; determines timing towards LGP-30 logic unit
						
		-- inside simulated LGP-30 
		
		-- mode switches
		bq, bqn: out std_logic;						-- manual input
		O1, O1n: out std_logic;						-- single step

		-- operation buttons
		bs: out std_logic;							-- start
		bcn: out std_logic;							-- clear counter
		br, brn: out std_logic;						-- fill instruction
		Ps, Psn: out std_logic;						-- execute instruction

		-- Latching buttons
		T: out std_logic_vector (1 to 4);		-- breakpoint switches
		D, Dn: out std_logic;						-- select 6 bit input
		Tc: out std_logic;							-- transfer control
		
		Tbs: in std_logic;							-- blocked state flag - drives stop/compute lights

		-- buttons on FlexoWriter in the real LGP-30
		readActive: in std_logic;					-- signals that FlexoWriter is ready to read (for read indicator led)
		readToggle: out std_logic;					-- alternate between ""start read" and "stop read"; actual status controlled by Flexowriter
		connected: out std_logic;					-- Flexowriter can send start signals to the logic unit
		
		-- Encoder for simulator control
		EncUp, EncDn: out std_logic;				-- pulses high for one io_clk
		EncPushed: out std_logic;					-- direct read of pushbutton;
		EncSelect: out std_logic;					-- toggled by encoder pushbutton
		
		-- option switches (toggles) on Mimas board
		option: out std_logic_vector (3 downto 0);
		
		-- external switches and LEDs
		switch: in std_logic_vector (24 downto 0);
		led: out std_logic_vector (22 downto 0)
	);
end panel;

architecture Behavioral of panel is

	-- timer counts down to ignore switch bounces within timer period.
	-- clocked by I/O clock
	constant deadtime: natural := 10000*CLK_IO_RATE;		-- debounce dead time 10 ms, in ioclock ticks
	signal deadtimer: natural range 0 to deadtime := 0;
	
	-- use same timer to determine pulse duration for pushbuttons
	-- clocked by CPU clock (to be compatible with variable CPU clock speed)
	constant pulsetime: natural := 125;		-- pulse time 1 ms @ 125 kHz CPU clock
	signal pulsetimer: natural range 0 to pulsetime := 0;
	signal pulse: std_logic := '0';

	-- copy of switch input vector after debouncing
	signal debounced: std_logic_vector (22 downto 0) := (others => '0');
	signal old : std_logic_vector (22 downto 0) := (others => '0');
	
	-- interlocked mode signals (mutually exclusive)
	signal mode_normal: std_logic := '0';
	signal mode_single: std_logic := '0';
	signal mode_manual: std_logic := '1';
		
	component toggle
		port (
			D: in std_logic;
			Q0, Q1, Q0n, Q1n: out std_logic
		);
	end component;
	
	component pushbutton
		port (
			D: in std_logic;				-- data input
			En: in std_logic;				-- enable
			Q, Qn: out std_logic;
			EnOut: out std_logic			-- output of enable status (for LED)
		);
	end component;
	
	component modebutton
		port (
			D: in std_logic;
			Q0, Q1, Qn: out std_logic
		);
	end component;

	component encoder is
		port (
			clk: 			in std_logic;
			X, Y: 		in std_logic;
			Qup, Qdn: 	out std_logic := '0'
		);
	end component;

begin

	------------------------------------------------
	-- component instances
	------------------------------------------------

	-- all toggles are initialized with the positive outputs set to '0'.
	-- connections to either positive or negative outputs, dependend on desired power-up state.
	
	T1_toggle: toggle PORT MAP (D => debounced(0), Q0 => T(1), Q1 => led (0), Q0n => open, Q1n => open);
	T2_toggle: toggle PORT MAP (D => debounced(1), Q0 => T(2), Q1 => led (1), Q0n => open, Q1n => open);
	T3_toggle: toggle PORT MAP (D => debounced(2), Q0 => T(3), Q1 => led (2), Q0n => open, Q1n => open);
	T4_toggle: toggle PORT MAP (D => debounced(3), Q0 => T(4), Q1 => led (3), Q0n => open, Q1n => open);
	D_toggle:  toggle PORT MAP (D => debounced(4), Q0 => D,    Q1 => led (4), Q0n => Dn, Q1n => open);		
	Tc_toggle: toggle PORT MAP (D => debounced(5), Q0 => Tc,   Q1 => led (5), Q0n => open, Q1n => open);

	-- read start/stop toggle controls readToggle output (to typewriter module);
	-- but its LED signals the actual tape reader state (controlled by input from typewriter unit)
	read_toggle: toggle PORT MAP (D => debounced(13), Q0 => readToggle,  Q1 => open, Q0n => open, Q1n => open);
	led(13) <= readActive;	
	
	conn_toggle:  toggle PORT MAP (D => debounced(14), Q0 => open, Q1 => open, Q0n => connected,  Q1n => led (14));
	Enc_toggle:   toggle PORT MAP (D => debounced(15), Q0 => EncSelect, Q1 => led(15), Q0n => open, Q1n => open);
	Spare_toggle: toggle PORT MAP (D => debounced(16), Q0 => open, Q1 => led (16), Q0n => open, Q1n => open);
	
	-- option toggles 
	-- on the landscape board, two of them are also operated via the switches in the "stop" and "comp" lights 
	opt0_toggle:  toggle PORT MAP (D => debounced(17) or debounced (21), Q0 => option(0), Q1 => led (17), Q0n => open, Q1n => open);	-- extended LCD - default OFF
	opt1_toggle:  toggle PORT MAP (D => debounced(18), Q0 => open, Q1 => open, Q0n => option(1), Q1n => led (18));							-- ESC sequences - default ON
	opt2_toggle:  toggle PORT MAP (D => debounced(19), Q0 => option(2), Q1 => led (19), Q0n => open, Q1n => open);							-- relay - default OFF 
	opt3_toggle:  toggle PORT MAP (D => debounced(20) or debounced (22), Q0 => open, Q1 => open, Q0n => option(3), Q1n => led (20));	-- extended HDMI - default ON
	
	start_pb: pushbutton PORT MAP (D => debounced (6) and pulse, En => '1', Q => bs, Qn => open, EnOut => led (6));
	clear_pb: pushbutton PORT MAP (D => debounced (7) and pulse, En => (not mode_normal), Q => open, Qn => bcn, EnOut => led (7));
	fill_pb:  pushbutton PORT MAP (D => debounced (8) and pulse, En => (not mode_normal), Q => br, Qn => brn, EnOut => led (8));
	exec_pb:  pushbutton PORT MAP (D => debounced (9) and pulse, En => mode_single, Q => Ps, Qn => Psn, EnOut => led (9));

	normal_mb: modebutton PORT MAP (D => mode_normal, Q0 => open, Q1 => led (10), Qn => open);
	single_mb: modebutton PORT MAP (D => mode_single, Q0 => O1, Q1 => led (11), Qn => O1n);
	manual_mb: modebutton PORT MAP (D => mode_manual, Q0 => bq, Q1 => led (12), Qn => bqn);
	
	-- Compute/Stop status indicators
	led (21) <= Tbs;		-- "compute" light
	led (22) <= not Tbs;	-- "stop" light
	
	-- Encoder
	encoder_enc: encoder PORT MAP (clk => io_clk, X => switch (24), Y => switch (23), Qup => EncUp, Qdn => EncDn);
 	
	-- poll switches with debounce timer
	process (io_clk) is
	begin 
		if rising_edge (io_clk) then
			
			-- Time the duration of output pulses to CPU.
			-- Clocked by CPU clock.
			if (lgp_clk = '1') then
				if (pulsetimer > 0) then 
					pulsetimer <= pulsetimer - 1;
					pulse <= '1';
				else 
					pulse <= '0'; 
				end if;
			end if;
			
			if (deadtimer = 0) then 
			
				-- outside of debounce dead time:
				-- whether the timer has expired, or a switch setting has changed (outside of debounce time):
				-- register new valid switch setting.
				-- invert switch polarity here - '1' in debounced denotes button is pressed.
				-- Encoder is excluded from this debouncing scheme!
				debounced <= not switch (22 downto 0);
				
				-- if switch settings have changed, restart debounce timer
				if switch (22 downto 0) /= old then
					old <= switch (22 downto 0);
					deadtimer <= deadtime;
					-- depending on CPU clock speed, pulses may take a long time and previous one may not be over yet
					if pulsetimer = 0 then pulsetimer <= pulsetime; end if;
				end if;
				
				-- manage interlocked switches
				if debounced (10) = '1' then
					mode_manual <= '0';
					mode_single <= '0';
					mode_normal <= '1';
				elsif debounced (11) = '1' then
					mode_manual <= '0';
					mode_single <= '1';
					mode_normal <= '0';
				elsif debounced (12) = '1' then
					mode_manual <= '1';
					mode_single <= '0';
					mode_normal <= '0';
				end if;
			else
				-- dead time still ongoing; ignore switch changes
				deadtimer <= deadtimer-1; 
			end if;
		end if;
	end process;
	
	EncPushed <= debounced(15);	-- direct encoder pushbutton output

end Behavioral;

