----------------------------------------------------------------------------------

-- LGP-30 drum computer simulation
--
-- Multiplex between UART and USB Serial communication,
-- depending on the setting of the external USB Select jumper.
-- Jumper present (usb_select_n = 0)  -> use USB.
--
--	Note that this unit does not control the UART RTS and CTS signals.
-- RTS output will still be driven (by the xmodem and flexowriter units) in USB mode.
-- CTS input (to xmodem unit) is clamped to '0' (clear to send) in USB mode, in lgp-30 main unit.
--
-- jm 1.8.2017

----------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;


entity usb_uart_mux is

    port(
		-- clocks 
		io_clk : in std_logic;						-- IO clock, 10 MHz
		usb_clk : in std_logic;						-- USB PHY clock, 60 MHz

		-- selector input
		usb_select_n: in std_logic;				-- 0=USB, 1=UART

		-- communication with internal units
		-- these are synchronized to io_clk
		data_tx:in std_logic_vector(7 downto 0);		
		data_rx: out std_logic_vector(7 downto 0);		
		write_strobe: in std_logic;				-- trigger 1 byte write to Tx buffer
		read_strobe: in std_logic;					-- trigger 1 byte read from Rx buffer
		tx_half_full: out std_logic; 				-- Tx buffer >= half full?
		rx_half_full: out std_logic; 				-- Rx buffer >= half full?
		rx_data_present: out std_logic; 			-- Rx buffer has character?
		
		-- external USB bus
		usb_dp : inout  STD_LOGIC;					-- Data+
		usb_dn : inout  STD_LOGIC;					-- Data-
		usb_led: out STD_LOGIC;						-- status LED
		
		-- external serial I/O
		uart_Rx: in std_logic;
		uart_Tx: out std_logic

	 );

end usb_uart_mux;

architecture Behavioral of usb_uart_mux is

	component uart
		port (
			io_clk: in std_logic;							-- I/O clock

			-- communication with internal units
			uart_data_tx:in std_logic_vector(7 downto 0);		
			uart_data_rx: out std_logic_vector(7 downto 0);		
			write_to_uart_tx: in std_logic;				-- trigger 1 byte write to Tx buffer
			read_from_uart_rx: in std_logic;				-- trigger 1 byte read from Rx buffer
			uart_tx_half_full: out std_logic; 			-- Tx buffer >= half full?
			uart_rx_half_full: out std_logic; 			-- Rx buffer >= half full?
			uart_rx_data_present: out std_logic; 		-- Rx buffer has character?
			
			-- external serial I/O
			uart_Rx: in std_logic;
			uart_Tx: out std_logic
		);
	end component;
	
	component usb is
		port(
			-- clocks 
			usb_clk : in std_logic;						-- USB PHY clock, 60 MHz

			-- communication with internal units
			-- these are synchronized to io_clk
			usb_data_tx:in std_logic_vector(7 downto 0);		
			usb_data_rx: out std_logic_vector(7 downto 0);		
			write_to_usb_tx: in std_logic;			-- trigger 1 byte write to Tx buffer
			read_from_usb_rx: in std_logic;			-- trigger 1 byte read from Rx buffer
			usb_tx_rdy: out std_logic; 				-- Tx buffer ready to accept byte?
			usb_rx_data_present: out std_logic; 	-- Rx buffer has character?
					
			-- external USB bus
			usb_dp : inout  STD_LOGIC;					-- Data+
			usb_dn : inout  STD_LOGIC;					-- Data-
			usb_led: out STD_LOGIC						-- status LED
		);
	end component;

	-- output signals -- multiplexed to module output depending on usb_select_n
	signal uart_tx_half_full, uart_rx_half_full, uart_rx_data_present: std_logic;
	signal usb_tx_rdy, usb_rx_data_present: std_logic;
	signal uart_data_rx, usb_data_rx: std_logic_vector(7 downto 0);

begin

	data_rx	<= usb_data_rx when (usb_select_n = '0') else uart_data_rx;									-- 8 bit Rx data 
	tx_half_full <= not usb_tx_rdy when (usb_select_n = '0') else uart_tx_half_full; 				-- Tx buffer >= half full?
	rx_half_full <= '0' when (usb_select_n = '0') else uart_rx_half_full; 								-- Rx buffer >= half full?
	rx_data_present <= usb_rx_data_present when (usb_select_n = '0') else uart_rx_data_present;	-- Rx buffer has character?

	inst_uart: uart 
		PORT MAP (
			io_clk		=> io_clk,

			-- communication with internal units
			uart_data_tx			=> data_tx,
			uart_data_rx			=> uart_data_rx,
			write_to_uart_tx 		=> write_strobe and usb_select_n,
			read_from_uart_rx		=> read_strobe and usb_select_n,
			uart_tx_half_full		=> uart_tx_half_full,
			uart_rx_half_full		=> uart_rx_half_full,
			uart_rx_data_present	=> uart_rx_data_present,
			
			-- external serial I/O
			uart_Rx		=> uart_Rx,
			uart_Tx		=> uart_Tx
		);
		
	inst_usb: usb 
		PORT MAP (
			usb_clk		=> usb_clk,
			
			-- communication with internal units
			-- these are synchronized to io_clk
			usb_data_tx				=> data_tx,		
			usb_data_rx 			=> usb_data_rx,
			write_to_usb_tx		=> write_strobe and not usb_select_n,	-- trigger 1 byte write to Tx buffer
			read_from_usb_rx		=> read_strobe and not usb_select_n,	-- trigger 1 byte read from Rx buffer
			usb_tx_rdy				=> usb_tx_rdy,									-- Tx buffer ready to accept byte?
			usb_rx_data_present	=> usb_rx_data_present, 					-- Rx buffer has character?

			-- external USB bus and status LED
			usb_dp		=> usb_dp,
			usb_dn		=> usb_dn,
			usb_led		=> usb_led
		);

end Behavioral;

