----------------------------------------------------------------------------------

-- LGP-30 drum computer simulation
-- XModem upload/download for magnetic drum images
--
-- Stores the main memory (64*64 words) and the A/R/C registers.
--
-- jm 10.9.2016
--
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library work;
use work.constants.all;


entity xmodem is
	port (
		-- begin/end xmodem mode
		xmodem_active: in std_logic;					-- xmodem transfer is active
		xmodem_done: out std_logic;					-- strobe high when we are done (or have failed)
	
		-- memory access to LGP-30 drum 
		-- system clock
		clk_tick: out std_logic;						-- can control clock from this unit via the single step mode
		-- main memory
		V: in std_logic;									-- read head
		Vp: out std_logic;								-- write head
		W: out std_logic;									-- write enable
		P: out std_logic_vector (1 to 6);			-- track selector
		sec_reset: out std_logic;						-- reset rotary position to sector 0
		
		-- registers
		A, R, C: in std_logic;							-- register read heads
		Ap, Rp, Cp: out std_logic;						-- register write heads - always on!
			
		-- connection to UART unit
		io_clk: in std_logic;							-- I/O clock
		uart_data_tx: out std_logic_vector(7 downto 0);		
		uart_data_rx: in std_logic_vector(7 downto 0);		
		write_to_uart_tx: out std_logic;				-- trigger 1 byte write to Tx buffer
		read_from_uart_rx: out std_logic;			-- trigger 1 byte read from Rx buffer
		uart_tx_half_full: in std_logic; 			-- Tx buffer >= half full?
		uart_rx_half_full: in std_logic; 			-- Rx buffer >= half full?
		uart_rx_data_present: in std_logic; 		-- Rx buffer has character?

		-- debug outputs
		test: out std_logic_vector (7 downto 0)
	);
end xmodem;

architecture Behavioral of xmodem is

	-- state control
	type control is 
	(
		idle,					-- no Rx or Tx is active
		
		rx_wait_new, 		-- wait for new block
		rx_wait_block,		-- wait for block number
		rx_wait_blockn,	-- wait for inverted block number
		rx_wait_data,		-- wait for 128 bytes of data
		rx_wait_check,		-- wait for checksum
		rx_done,				-- have sent final ACK (for EOT). Nothing more to do except to reset uart read strobe 
		rx_delay,
		rx_store,			-- store complete block
		rx_clear,			-- wait for data buffer to clear (error recvery)
		
		tx_new,				-- digest ACK/NAK and send new block
		tx_send_block,		-- send block number
		tx_send_blockn,	-- send inverted block number
		tx_send_data,		-- send 128 bytes of data
		tx_send_check,		-- send checksum
		tx_send_eot,		-- send eot
		tx_wait_ack,		-- wait for ACK (either for data block or for EOT)
		tx_delay,			-- delay for processing of pulses (setup) before loading from drum
		tx_load				-- load new block from drum
	);
	signal state: control := idle;									-- main state machine
	
	-- XModem protocol character constants
	constant NAK: std_logic_vector (7 downto 0) := "00010101";
	constant ACK: std_logic_vector (7 downto 0) := "00000110";
	constant EOT: std_logic_vector (7 downto 0) := "00000100";
	constant SOH: std_logic_vector (7 downto 0) := "00000001";
	constant CAN: std_logic_vector (7 downto 0) := "00011000";
	
	-- Timeout monitor 
	constant timeout_1s: natural := CLK_IO_RATE*1000000;
	constant timeout_10s: natural := CLK_IO_RATE*1000000*10;
	signal timer: natural range 0 to timeout_10s:= 0;
	
	-- Error count
	constant NAK_max: natural := 9;
	signal NAK_count: natural range 0 to NAK_max := 0;
	
	-- Xmodem protocol variables
	signal due_block: natural range 0 to 255;						-- block number (expected/to be sent)
	signal this_block: natural range 0 to 255;					-- block number (received)
	signal bytecount: natural range 0 to 127;						-- byte count 
	signal checksum: natural range 0 to 255;						-- checksum for xmodem blocks
	
	-- strobes to other units
	signal write_strobe: std_logic := '0';							-- readable copy of write_to_uart_tx
	signal read_strobe: std_logic := '0';							-- readable copy of read_from_uart_rx
	signal clk_tick_int: std_logic := '0';							-- readable copy of clk_tick
	
	-- Rx/Tx buffer management
	signal buf_bit: natural range 0 to 7;							-- bit within buffer byte
	signal buf_byte: natural range 0 to 127;						-- byte within buffer
	signal buf_block: natural range 0 to 127;						-- block counter: 6 bit track + half-track
	
	type buf_type is array (0 to 127) of std_logic_vector (7 downto 0);
	signal rx_buf, tx_buf: buf_type;									
	-- Note: two  separate buffer synthesize MUCH more efficiently than a shared buffer for Rx/Tx 
	-- (although only one is used at any time). Avoids a large number of tri-state buffers.

begin

	process (io_clk)
	
		variable rx_byte, tx_byte: std_logic_vector (7 downto 0);			
		-- intermediate buffer for rx_buf, tx_buf bytes 
		-- (allows use of 8 bit block RAM -- more effective design, faster synthesis)

		procedure send_ack (newstate: in control) is
		begin
			uart_data_tx <= ACK;
			write_strobe <= '1';
			state <= newstate;
			NAK_count <= 0;
		end;
		
	begin
		if rising_edge (io_clk) then
		
			-- reset strobe signal that signals completion
			xmodem_done <= '0';
			
			if xmodem_active = '1' then

				-- Reset strobe signals
				-- These can be overwritten by setting the strobes further down
				read_strobe <= '0';
				write_strobe <= '0';
				clk_tick_int <= '0';
				
				-- count down timer
				if (timer > 0) then timer <= timer-1; end if;

				------------------------------
				-- XModem state machine
				------------------------------

				case state is
				
					------------------------------
					-- start from Idle mode
					------------------------------
					when idle =>
						if (uart_rx_data_present = '1') then
							if (uart_data_rx = xmodem_start_rx) then		-- start of reception requested
								state <= rx_delay;								-- delay one tick, then receive a block
								read_strobe <= '1';								-- digest character
								sec_reset <= '1';									-- reset drum position (async.)
								buf_byte <= 0;
								buf_bit <= 0;
								buf_block <= 0;									-- reset block count (track# plus half-track)
								due_block <= 1;
								NAK_count <= 0;
							elsif (uart_data_rx = xmodem_start_tx) then	-- start of transmission requested
								state <= tx_delay;								-- delay for processing of pulses, then load first block from the drum
								read_strobe <= '1';								-- digest character
								sec_reset <= '1';									-- reset drum position
								buf_block <= 0;									-- reset block count (track# plus half-track)
								buf_byte <= 0;
								buf_bit <= 0;
								due_block <= 1;
							else
								xmodem_done <= '1'; 								-- false alarm
							end if;
						else
							xmodem_done <= '1'; 									-- false alarm
						end if;
					
					------------------------------
					-- receive state machine
					------------------------------
					
					when rx_wait_new to rx_wait_check =>
						-- these character reception states share a common behavior
												
						if (uart_rx_data_present = '1') and (read_strobe = '0') then
							timer <= timeout_1s;
							read_strobe <= '1';									-- digest character

							case state is
							
								when rx_wait_new =>
									if uart_data_rx = SOH then					-- block starts
										state <= rx_wait_block;
									elsif uart_data_rx = EOT then				-- transmission complete
										send_ack (rx_done);	
									else
										timer <= timeout_1s;
										state <= rx_clear;
									end if;

								when rx_wait_block =>
									if to_integer(unsigned(uart_data_rx)) <= due_block then		-- expect block number
										this_block <= to_integer(unsigned(uart_data_rx));			-- smaller numbers (repeats) are tolerated but ignored
										state <= rx_wait_blockn;
									else
										timer <= timeout_1s;
										state <= rx_clear;
									end if;
								
								when rx_wait_blockn =>
									if to_integer(unsigned(uart_data_rx)) = 255-this_block then		-- expect inverted block number
										checksum <= 0;
										bytecount <= 0;
										state <= rx_wait_data;
									else
										timer <= timeout_1s;
										state <= rx_clear;
									end if;
								
								when rx_wait_data =>
									rx_buf (bytecount) <= uart_data_rx;
									checksum <= checksum + to_integer(unsigned(uart_data_rx));
									if bytecount = 127 then
										state <= rx_wait_check;
									end if;
									bytecount <= bytecount+1;

								when rx_wait_check =>
									if to_integer(unsigned(uart_data_rx)) = checksum then			-- correct checksum
										if this_block = due_block then		-- have received the next block due
											due_block <= due_block+1;			-- increment after successful reception
											state <= rx_store;					-- write data to drum
										else											-- repeat of a block already received
											send_ack (rx_wait_new);				-- acknowledge, but ignore 
										end if;
									else
										timer <= timeout_1s;
										state <= rx_clear;
									end if;
							
								when others => 									-- should be redundant							
							end case;
						
						elsif timer=0 then										-- reception has timed out: send NAK
							timer <= timeout_1s;
							state <= rx_clear;
						end if;
					
					when rx_clear =>												-- NAK: wait for Rx buffer to clear, then send NAK
						if (uart_rx_data_present = '1') and (read_strobe = '0') then
							-- flush buffer
							timer <= timeout_1s;
							read_strobe <= '1';									-- digest character
						elsif timer = 0 then
							-- once buffer is empty, send NAK
							uart_data_tx <= NAK;
							if NAK_count = NAK_max then						-- failed: cancel transmission
								uart_data_tx <= CAN;							
								state <= rx_done;
							else
								NAK_count <= NAK_count+1;
								timer <= timeout_10s;
								state <= rx_wait_new;
							end if;
							write_strobe <= '1';
						end if;
					
					when rx_delay => 
						sec_reset <= '0';											-- release reset
						timer <= timeout_10s;
						state <= rx_wait_new;									-- 1 tick delay to get addresses and data set up
					
					when rx_done =>												-- have ACK'd EOT, or CANceled - just need to clean up
						W <= '0';													-- (extra state is needed to ensure that uart write strobe is cleared)
						xmodem_done <= '1';
						state <= idle;

					------------------------------
					-- transmit state machine
					------------------------------

					when tx_new to tx_send_eot => 
					
						-- check if receiver has sent NAK.
						-- ignore any other characters
						if (uart_rx_data_present = '1') then
							read_strobe <= '1';									-- digest character
							if uart_data_rx = NAK then
								state <= tx_new;									-- retransmit
							end if;
						
						-- Only transmit when our Tx buffer has room and the receiver is ready; 
						-- and allow the uart write strobe to clear(do not transmit in two consecutive cycles).
						-- Note counter-intuitive polarity of uart_CTS...
						elsif (uart_tx_half_full = '0') and (write_strobe = '0') then
							case state is
							
								when tx_new => 
									uart_data_tx <= SOH;
									state <= tx_send_block;
									
								when tx_send_block =>
									uart_data_tx <= std_logic_vector (to_unsigned (due_block, 8));
									state <= tx_send_blockn;
									
								when tx_send_blockn =>
									uart_data_tx <= std_logic_vector (to_unsigned (255-due_block, 8));
									checksum <= 0;
									bytecount <= 0;
									state <= tx_send_data;
									
								when tx_send_data =>
									uart_data_tx <= tx_buf (bytecount);
									checksum <= checksum + to_integer(unsigned(tx_buf (bytecount)));
									if bytecount = 127 then
										state <= tx_send_check;
									end if;
									bytecount <= bytecount+1;
									
								when tx_send_check => 
									uart_data_tx <= std_logic_vector (to_unsigned (checksum, 8));
									timer <= timeout_10s;
									state <= tx_wait_ack;
							
								when tx_send_eot => 
									uart_data_tx <= EOT;
									timer <= timeout_10s;
									state <= tx_wait_ack;

								when others => 									-- should be redundant

							end case;
							write_strobe <= '1';
						end if;
					
					when tx_delay => 
						sec_reset <= '0';											-- release reset
						state <= tx_load;											-- 1 tick delay to get addresses and data set up

					when tx_wait_ack =>
						if (uart_rx_data_present = '1') then
							read_strobe <= '1';									-- digest character
							if uart_data_rx = ACK then
								case due_block is
									when 128+16 => 								-- last block was ACK'd, send EOT
										state <= tx_send_eot;		
									when 128+16+1 => 								-- EOT was ACK'd, we are done
										xmodem_done <= '1'; 
										state <= idle;					
									when others => 								-- more blocks to send
										state <= tx_load;							-- load new data
								end case;
								due_block <= due_block+1;
							elsif uart_data_rx = NAK then
								state <= tx_new;									-- resend block
							end if;		
						elsif timer=0 then
							xmodem_done <= '1'; 
							state <= idle;											-- give up -- introduce retries here? ###
						end if;
					
					-----------------------------------
					-- read/write LGP-30 drum 
					-----------------------------------

					when rx_store =>
						if clk_tick_int = '0' then								-- previous tick is over
							clk_tick_int <= '1';			
							
							if due_block <= 128 then							-- main memory block
								W <= '1';											-- write to drum
								Ap <= A; Rp <= R; Cp <= C;						-- always recycle all registers
								if buf_bit=0 then rx_byte := rx_buf (buf_byte); end if;
								Vp <= rx_byte (buf_bit);
								if buf_bit=7 then
									if buf_byte=127 then
										buf_block <= buf_block+1;
										send_ack (rx_wait_new);					-- ready to read new block
									end if;
									buf_byte <= buf_byte+1;
								end if;
								buf_bit <= buf_bit+1;
							else														-- A/R/C register block
								W <= '0';											-- no more writing to data memory
								rx_byte := rx_buf (buf_byte); 
								Ap <= rx_byte (2);
								Rp <= rx_byte (1);
								Cp <= rx_byte (0);
								if buf_byte=127 then
									send_ack (rx_wait_new);						-- ready to read new block
								end if;
								buf_byte <= buf_byte+1;
							end if;
						end if;

					when tx_load =>
						W <= '0';													-- read from drum
						if clk_tick_int = '0' then								-- previous tick is over
							clk_tick_int <= '1';	
							Ap <= A; Rp <= R; Cp <= C;							-- always recycle all registers

							if due_block <= 128 then							-- main memory block
								tx_byte (buf_bit) := V;
								if buf_bit=7 then
									tx_buf (buf_byte) <= tx_byte;
									if buf_byte=127 then
										buf_block <= buf_block+1;
										state <= tx_new;							-- transmit this block
									end if;
									buf_byte <= buf_byte+1;
								end if;
								buf_bit <= buf_bit+1;
							else														-- A/R/C register block
								tx_buf (buf_byte) <= "00000" & A & R & C;
								if buf_byte=127 then
									state <= tx_new;								-- transmit this block
								end if;
								buf_byte <= buf_byte+1;
							end if;

						end if;
										
				end case;
			end if;
			
			-- set test outputs for debugging
			if state=rx_clear 		then test (1) <= '1'; else test(1) <= '0'; end if;
			if state=rx_wait_check 	then test (3) <= '1'; else test(3) <= '0'; end if;
			if state=rx_wait_data 	then test (5) <= '1'; else test(5) <= '0'; end if;
			if state=rx_store 		then test (7) <= '1'; else test(7) <= '0'; end if;
			test (0) <= '0';
			test (2) <= '0';
			test (4) <= '0';
			test (6) <= '0';
	
		end if;
	end process;
	
	
	-- external output for read and write strobes
	write_to_uart_tx <= write_strobe;
	read_from_uart_rx <= read_strobe;
	
	-- external output for drum clock
	clk_tick <= clk_tick_int;

	-- map 128 byte block counter to drum track counter
	P <= std_logic_vector (to_unsigned (buf_block/2, 6));

end Behavioral;

