----------------------------------------------------------------------------------

-- LGP-30 drum computer simulation:
-- USB Serial Interface (used by flexowriter and xmodem).
-- Encapsulate USB Serial core to make the interface similar to the UART unit.
--
-- jm 30.7.2017
--
-- Underlying USB modules are 3rd party open source:
--
-- USB Serial Data Transfer Core
-- Version: 2011-10-04
-- Author: Joris van Rantwijk
-- Language: VHDL
-- License: GPL - GNU General Public License
-- Website: http://jorisvr.nl/usb/
--
-- USB 1.1 PHY Layer
-- Author: Rudolf Usselmann, rudi@asics.ws
-- Downloaded from: http://www.opencores.org/cores/usb_phy/
-- Copyright (C) 2000-2002 Rudolf Usselmann, www.asics.ws, rudi@asics.ws
-- Verilog to VHDL conversion by Martin Neumann martin@neumnns-mail.de

----------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

use work.usb_pkg.all;
use work.usb_phy_pkg.all;

entity usb is

    port(
		-- clocks 
		usb_clk : in std_logic;						-- USB PHY clock, 60 MHz

		-- communication with internal units
		-- these are synchronized to io_clk
		usb_data_tx:in std_logic_vector(7 downto 0);		
		usb_data_rx: out std_logic_vector(7 downto 0);		
		write_to_usb_tx: in std_logic;			-- trigger 1 byte write to Tx buffer
		read_from_usb_rx: in std_logic;			-- trigger 1 byte read from Rx buffer
		usb_tx_rdy: out std_logic; 				-- Tx buffer ready to accept byte?
		usb_rx_data_present: buffer std_logic; -- Rx buffer has character?
		
		-- external USB bus
		usb_dp : inout  STD_LOGIC;					-- Data+
		usb_dn : inout  STD_LOGIC;					-- Data-
		usb_led: out STD_LOGIC						-- status LED
	 );

end entity usb;


architecture Behavioral of usb is

    constant RXBUFSIZE_BITS : integer := 9;			-- ### this buffer size determines RX behavior/overflow
    constant TXBUFSIZE_BITS : integer := 9;

	 -- internal Transceiver interface
	 signal xcvr_rxd, xcvr_rxdp, xcvr_rxdn: std_logic;
	 signal xcvr_txdp, xcvr_txdn, xcvr_txoe: std_logic;

	 -- internal UTMI Interface
    signal PHY_DATAOUT :     std_logic_vector(7 downto 0);
    signal PHY_TXVALID :		std_logic;
    signal PHY_TXREADY :     std_logic;
    signal PHY_DATAIN :      std_logic_vector(7 downto 0);
    signal PHY_RXVALID :     std_logic;
    signal PHY_RXACTIVE :    std_logic;
    signal PHY_RXERROR :     std_logic;
    signal PHY_LINESTATE :   std_logic_vector(1 downto 0);

    -- USB Serial interface signals (high level)
    signal usb_devreset :   std_logic;
    signal usb_busreset :   std_logic;
    signal usb_highspeed :  std_logic;
    signal usb_suspend :    std_logic;
    signal usb_online :     std_logic;
    signal usb_rxval :      std_logic;
    signal usb_rxdat :      std_logic_vector(7 downto 0);
    signal usb_rxrdy :      std_logic;
    signal usb_rxlen :      std_logic_vector(RXBUFSIZE_BITS-1 downto 0);
    signal usb_txval :      std_logic;
    signal usb_txdat :      std_logic_vector(7 downto 0);
    signal usb_txrdy :      std_logic := '1';
    signal usb_txroom :     std_logic_vector(TXBUFSIZE_BITS-1 downto 0);
    signal s_txcork : 		 std_logic := '0';

	 -- unused outputs of USB_Serial unit
	 -- (not present on simplified USB 1.1 PHY layer?)
	signal PHY_RESET :       std_logic; -- reset the PHY layer - active high
	signal PHY_XCVRSELECT :  std_logic;	-- high speed vs. full speed -- can only use full speed here
	signal PHY_TERMSELECT :  std_logic; -- high speed vs. full speed termination -- can only use full speed here
	signal PHY_OPMODE :      std_logic_vector(1 downto 0);
	 
    -- General purpose counter
    signal clkcnt : unsigned(24 downto 0);
	 
	 -- sync between slow io_clk drive strobes and fast usb_clk
	 signal write_ongoing: std_logic := '0';

	component usb_xcvr is
		Port ( 
		 	-- external USB bus
			usb_dp : inout  STD_LOGIC;				-- Data+
			usb_dn : inout  STD_LOGIC;				-- Data-
				  
			-- internal I/O and control
			rxd, rxdp, rxdn  : out  std_logic;
			txdp, txdn, txoe : in std_logic;
			
			mode : in STD_LOGIC := '1'				-- high for differential signaling, low for txdn = FSEO
		);
	end component;

begin

	-- Transceiver (J. Mueller)
	usb_xcvr_inst: usb_xcvr
	port map (
		-- external USB bus
		usb_dp => usb_dp,							-- Data+
		usb_dn => usb_dn,							-- Data-
			  
		-- internal I/O and control
		rxd => xcvr_rxd, 
		rxdp => xcvr_rxdp, 
		rxdn => xcvr_rxdn,
		txdp => xcvr_txdp, 
		txdn => xcvr_txdn, 
		txoe => xcvr_txoe,
		mode => '1'
	);

	-- UTMI (R. Usselmann)
	usb_phy_inst: usb_phy 
		generic map (
			usb_rst_det => TRUE
		)
		port map (
			 clk              => usb_clk,		  	-- 60 MHz
			 rst              => not PHY_RESET,	-- serial device can reset the PHY layer. rst is active low! 
			 phy_tx_mode      => '1',				-- HIGH level for differential io mode (else single-ended)
			 usb_rst          => open,				-- reset detected on USB (long SE0) -- nobody cares? monitored by Serial Init as well?

			 -- external Transceiver Interface
			 rxd => xcvr_rxd, 
			 rxdp => xcvr_rxdp, 
			 rxdn => xcvr_rxdn,
			 txdp => xcvr_txdp, 
			 txdn => xcvr_txdn, 
			 txoe => xcvr_txoe,						-- active low

			 -- UTMI Interface
			 DataOut_i => PHY_DATAOUT,
			 TxValid_i => PHY_TXVALID,
			 TxReady_o => PHY_TXREADY,
			 DataIn_o => PHY_DATAIN,
			 RxValid_o => PHY_RXVALID,
			 RxActive_o => PHY_RXACTIVE,
			 RxError_o => PHY_RXERROR,
			 LineState_o => PHY_LINESTATE
		);


    -- Direct interface to serial data transfer component
    usb_serial_inst : usb_serial
        generic map (
				-- using generic vendor and product IDs from Objective Development:
				-- https://github.com/obdev/v-usb/blob/master/usbdrv/USB-IDs-for-free.txt
				-- jm 29.9.17
            VENDORID        => X"16c0",
            PRODUCTID       => X"05e1",
            VERSIONBCD      => X"0031",
            VENDORSTR       => "jm@e-basteln.de",
            PRODUCTSTR      => "LittleGP-30",
            SERIALSTR       => "20170806",
--            VENDORSTR       => "Joris",
--            PRODUCTSTR      => "Trenz TE0146 module with fpga_usb_serial firmware",
--            SERIALSTR       => "20111004",
            HSSUPPORT       => false,				-- jm 6.8.17 - PHY layer does not support High Speed
            SELFPOWERED     => false,
            RXBUFSIZE_BITS  => RXBUFSIZE_BITS,
            TXBUFSIZE_BITS  => TXBUFSIZE_BITS )
        port map (
            CLK             => usb_clk,
            RESET           => usb_devreset,
            USBRST          => usb_busreset,
            HIGHSPEED       => usb_highspeed,
            SUSPEND         => usb_suspend,
            ONLINE          => usb_online,
            RXVAL           => usb_rxval,
            RXDAT           => usb_rxdat,
            RXRDY           => usb_rxrdy,
            RXLEN           => usb_rxlen,
            TXVAL           => usb_txval,
            TXDAT           => usb_txdat,
            TXRDY           => usb_txrdy,
            TXROOM          => usb_txroom,
            TXCORK          => s_txcork,
            PHY_DATAIN      => PHY_DATAIN,
            PHY_DATAOUT     => PHY_DATAOUT,
            PHY_TXVALID     => PHY_TXVALID,
            PHY_TXREADY     => PHY_TXREADY,
            PHY_RXACTIVE    => PHY_RXACTIVE,
            PHY_RXVALID     => PHY_RXVALID,
            PHY_RXERROR     => PHY_RXERROR,
            PHY_LINESTATE   => PHY_LINESTATE,
            PHY_OPMODE      => PHY_OPMODE,
            PHY_XCVRSELECT  => PHY_XCVRSELECT,
            PHY_TERMSELECT  => PHY_TERMSELECT,
            PHY_RESET       => PHY_RESET );


    -- Show USB status on LED
    usb_led <= '0' when (
        -- suspended -> on
        (usb_suspend = '1') or
        -- offline -> blink 8%
        (usb_online = '0' and clkcnt(24 downto 22) = "000") or
        -- highspeed -> fast blink 50%
        (usb_online = '1' and usb_highspeed = '1' and clkcnt(22) = '0') or
        -- fullspeed -> slow blink 50%
        (usb_online = '1' and usb_highspeed = '0' and clkcnt(24) = '0') ) else '1';

    -- Assign signals to USB component
    usb_devreset <= '0';

	-- interact with LGP-30.
	usb_tx_rdy <= usb_txrdy;
	
	process (usb_clk) is
	begin
		if rising_edge (usb_clk) then
		
			-- Clock counter
			clkcnt <= clkcnt + 1;

			-- write to USB
			
			usb_txval <= '0';		-- reset TX burst signal
			if write_to_usb_tx = '1' and usb_online = '1' and usb_txrdy = '1' and write_ongoing = '0' then		
				-- Host has new byte to write, USB is ready.
				-- Note that write_to_usb_tx strobe will be long (10 MHz clock).
				-- Send only one byte per tx strobe; use write_ongoing to track ongoing strobe.
				usb_txdat <= usb_data_tx;
				usb_txval <= '1';
				write_ongoing <= '1';	-- no further writes while the write strobe remains high
			elsif write_to_usb_tx = '0' then
				-- host write strobe ended
				write_ongoing <= '0';
			end if;
			
			-- read from USB:
			-- usb_rxrdy normally remains high (we are ready to receive a byte)
			
			if usb_rx_data_present = '0' and usb_online = '1' and usb_rxrdy = '1' and usb_rxval = '1' then
				-- New byte in USB FIFO, and our output buffer is empty
				usb_data_rx <= usb_rxdat;		-- get byte from FIFO
				usb_rx_data_present <= '1';	-- signal to host that data are ready
				usb_rxrdy <= '0';					-- FIFO must push no further data

			elsif usb_rx_data_present = '1' and read_from_usb_rx = '1' then
				-- the host is reading the current byte from usb_data_rx
				usb_rx_data_present <= '0';	-- to host: no more data for now

			elsif usb_rx_data_present = '0' and usb_online = '1' and read_from_usb_rx = '0' then
				-- host is done reading;
				-- ready to receive the next byte from the USB FIFO
				usb_rxrdy <= '1';
			end if;
			
		end if;
	end process;

end architecture Behavioral;

